#ifndef __HEXDUMPER_H__
#define __HEXDUMPER_H__

#include <vector>
// object to output very configurable hexdumps.
class HexDumper {
public:
    const size_type npos= (size_type)-1;
    enum DumpUnitType {
        DUMPUNIT_BYTE=0,
        DUMPUNIT_WORD=1,
        DUMPUNIT_DWORD=2
    };
    enum DumpFormat {
        DUMP_HEX_ASCII,     // [offset] 50 51 52 53 0d 0a 00 01  PQRS....
        DUMP_HEX,           // [offset] 50 51 52 53 0d 0a 00 01
        DUMP_ASCII,         // [offset] PQRS....
        DUMP_STRINGS,       // [offset] "PQRS",0d,0a,00,01
        DUMP_RAW,           // [offset] PQRS^M^J^@^A
        DUMP_MD5            // just print md5 of entire buffer
    };

    HexDumper(size_t ofs=0, DumpUnitType t=DUMPUNIT_BYTE, DumpFormat f=DUMP_HEX_ASCII, bool withoffset=true, bool summarize=true)
        : m_offset(ofs), m_dumpunit(t), m_dumpformat(f), m_bWithOffset(withoffset), m_bSummarize(summarize)
        {
        }
    ~HexDumper();

    // sets current offset
    HexDumper& offset(size_t ofs) { m_offset= ofs; }

    // sets nr of items printed per line
    // not applicable for 'strings', and 'raw' output
    // 0 means all on one line.
    HexDumper& width(size_t width) { m_width= width; }

    // sets dump units
    HexDumper& unit(DumpUnitType t) { m_dumpunit= t; }
    HexDumper& byte() { m_dumpunit= DUMPUNIT_BYTE; }
    HexDumper& word() { m_dumpunit= DUMPUNIT_WORD; }
    HexDumper& dword() { m_dumpunit= DUMPUNIT_DWORD; }

    // sets type of output
    HexDumper& format(DumpFormat t) { m_dumpformat= t; }
    HexDumper& hexascii() { m_dumpformat= DUMP_HEX_ASCII   ; }
    HexDumper& hex()      { m_dumpformat= DUMP_HEX         ; }
    HexDumper& ascii()    { m_dumpformat= DUMP_ASCII       ; }
    HexDumper& strings()  { m_dumpformat= DUMP_STRINGS     ; }
    HexDumper& raw()      { m_dumpformat= DUMP_RAW         ; }
    HexDumper& md5()      { m_dumpformat= DUMP_MD5         ; }
    HexDumper& withoffset(bool flag=true) { m_bWithOffset= flag; }
    HexDumper& summarize(bool flag=true) { m_bSummarize= flag; }

    template<T>
    void dump(const std::vector<T>& buf, size_t ofs=0, size_t count=npos);
    template<T>
    void dump(const std::vector<T>::const_iterator& start, const std::vector<T>::const_iterator& end);
    void dump(const void*buf, size_t len);
    void end();

    // retrieve current output buffer
    std::string getstring();

private:
    typedef std::pair<size_t,std::string> lime_item_t;
    std::vector<line_item_t> m_out;
    bool m_bWithOffset;
    bool m_bSummarize;
    DumpFormat m_dumpformat;
    DumpUnitType m_dumpunit;
    size_t m_offset;
    size_t m_width;
};

template<T>
void HexDumper::dump(const std::vector<T>& buf, size_t ofs=0, size_t count=npos)
{
    dump((const void*)&buf[ofs], count*sizeof(T));
}

template<T>
void HexDumper::dump(const std::vector<T>::const_iterator& start, const std::vector<T>::const_iterator& end)
{
    dump((const void*)&(*start), (end-start)*sizeof(T));
}
#endif
