/* (C) 2003 XDA Developers  itsme@xs4all.nl
 *
 * $Header: /var/db/cvs/xda-devtools/itsutils/common/stringutils.h,v 1.7 2005/06/12 22:52:04 itsme Exp $
 */

#ifndef __STRINGUTILS_H_

// make sure everything is compiled with '_MT' defined.
// otherwise this will not link with code that has any of the mfc
// headers included.
//    without _MT this code compiles to non multithreaded templates,
//    with it compiles to templates with threading support.

#include <tchar.h>

#include <string>
#include <vector>
#include "vectorutils.h"

typedef std::vector<std::string> StringList;

namespace std {
#ifdef _UNICODE
    typedef wstring tstring;
#else
    typedef string tstring;
#endif
}

//-------------------------------------------------------
//  string conversion functions
//
std::string ToString(WCHAR* tbuf);
std::wstring ToWString(WCHAR* tbuf);
std::tstring ToTString(WCHAR* tbuf);

std::string ToString(char* buf);
std::wstring ToWString(char* buf);
std::tstring ToTString(char* buf);

std::string  ToString(const std::string& str);
std::wstring ToWString(const std::string& str);
std::tstring ToTString(const std::string& str);

std::string  ToString(const std::wstring& wstr);
std::wstring ToWString(const std::wstring& str);
std::tstring ToTString(const std::wstring& str);

/*
std::string ToString(TCHAR* tbuf);
std::wstring ToWString(TCHAR* tbuf);
std::tstring ToTString(TCHAR* tbuf);

std::string  ToString(const std::tstring& str);
std::wstring ToWString(const std::tstring& str);
//std::tstring ToTString(const std::tstring& str);
*/

//-------------------------------------------------------
//  string manipulation functions
//
void chomp(std::string& str);
void chomp(char *str);
bool SplitString(const std::string& str, StringList& strlist, bool bWithEscape= true);
std::string JoinStringList(const StringList& strlist, const std::string& sep);

std::string stringformat(char *fmt, ...);

int stringicompare(const std::string& a, const std::string& b);
std::string tolower(const std::string& str);

#define stringptr(v)  (&(v)[0])

std::string hexdump(const ByteVector& buf, int nDumpUnitSize=1);
std::string hexdump(const BYTE *buf, int nLength, int nDumpUnitSize=1);
std::string hexdump(DWORD dwOffset, const BYTE *buf, int nLength, int nDumpUnitSize=1, int nMaxUnitsPerLine=16);

std::string ascdump(const ByteVector& buf, const std::string& escaped= "", bool bBreakOnEol= false);
std::string asciidump(const BYTE *buf, size_t bytelen);

std::string hash_as_string(const ByteVector& hash);

#define __STRINGUTILS_H_
#endif
