/* (C) 2003 XDA Developers  itsme@xs4all.nl
 *
 * $Header: /var/db/cvs/xda-devtools/itsutils/dll/itsutils.cpp,v 1.34 2005/06/23 11:34:25 itsme Exp $
 */
#include <windows.h>
#include <Tlhelp32.h>

// !!! I am mixing up handles and processid's quite often.
//     on WinCE this is not a problem, but this needs to be
//     fixed before porting to another platform.

// todo:
//    * merge docread and sdread and p300read
//    * add options to specify 
//        - ioctl read blocksize ( how large blocks are read at once. )
//        - sectorsize ( for converting offset -> sector nr )
//        - ITReadDisk blocksize ( how large blocks are transfered over rapi at once )
//        - file blocksize ( how large blocks are read/written at once )
#include "ItsUtils.h"
#include "dllversion.h"

#include "kernelinfo.h"

#include "WINDEV.H"
#include "diskio.h"
#include "cenk.h"

#include "GsmDevice.h"
#include "debug.h"
#include "ptrutils.h"

#define min(a,b) ((a)<(b)?(a):(b))

#define AT_NONEXISTANT      1
#define AT_ISDIRECTORY      2
#define AT_ISFILE           3
int getWin32Attributes(const TCHAR *szName)
{
    WIN32_FIND_DATA wfd;
    HANDLE hFind = FindFirstFile( szName, &wfd);
    if (INVALID_HANDLE_VALUE == hFind)
        return AT_NONEXISTANT;

    FindClose( hFind);

    if (wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
        return AT_ISDIRECTORY;

    return AT_ISFILE;
}

BOOL APIENTRY DllMain( HANDLE hModule,
                       DWORD  ul_reason_for_call,
                       LPVOID lpReserved
                     )
{
    DebugOutputDebugString();

    DebugSetLogfile("itsutils.log");

    switch (ul_reason_for_call)
    {
        case DLL_PROCESS_ATTACH: //debug("DLL_PROCESS_ATTACH\n"); break;
        case DLL_THREAD_ATTACH:  //debug("DLL_THREAD_ATTACH\n"); break;
        case DLL_THREAD_DETACH:  //debug("DLL_THREAD_DETACH\n"); break;
        case DLL_PROCESS_DETACH: //debug("DLL_PROCESS_DETACH\n"); break;
            ;
    }
    return TRUE;
}

void logtime(char *msg)
{
    static DWORD tprev=0;
    DWORD t= GetTickCount();

    if (tprev)
        debug("(%4d) %8d : %s\n", t-tprev, t, msg);
    else
        debug("       %8d : %s\n", t, msg);

    tprev= t;
}
bool isWince3x()
{
    OSVERSIONINFO version;
    if (!GetVersionEx(&version)) {
        error("GetVersionEx");
        return false;
    }
    return (version.dwMajorVersion==3);
}

ITSUTILS_API HRESULT STDAPICALLTYPE ITGetVersion(
        DWORD cbInput, void *pbInput,
        DWORD *pcbOutput, void **ppbOutput,
        IRAPIStream *pStream)
{
    *pcbOutput=0;
    *ppbOutput=NULL;

    return ITSDLL_VERSION;
}

// returns 0 when process not found, 1 when process not killed, 2 when process killed.
ITSUTILS_API HRESULT STDAPICALLTYPE ITTerminateProcess(
        DWORD cbInput, TerminateProcessParams *pbInput,
        DWORD *pcbOutput, BYTE **ppbOutput,
        IRAPIStream *pStream)
{
    *ppbOutput=NULL;
    *pcbOutput=0;


    HANDLE hProc= pbInput->dwHandle
            ?(HANDLE)pbInput->dwHandle
            :GetProcessHandle(pbInput->wszProcessName);

    if (hProc==NULL || hProc==INVALID_HANDLE_VALUE) {
        return ERROR_NOT_FOUND;
    }

    DWORD res= 0;
    if (!TerminateProcess(hProc, 0))
        res= GetLastError();
    CloseHandle(hProc);

    return res;
}

#define MEM_PAGE_SIZE 4096

bool ReadMemory(DWORD dwStart, BYTE *buf, DWORD dwLength, DWORD *pdwCopied)
{
    DWORD nCopied= 0;
    DWORD addr= dwStart;

    while (nCopied < dwLength)
    {
        DWORD nChunkSize= min(MEM_PAGE_SIZE-(addr&(MEM_PAGE_SIZE-1)), dwLength-nCopied);

        if (IsBadReadPtr((void*)addr, nChunkSize))
            break;
        memcpy(&buf[nCopied], (void*)addr, nChunkSize);
        nCopied += nChunkSize;
        addr += nChunkSize;
    }
    *pdwCopied= nCopied;

    return true;
}

bool WriteMemory(DWORD dwStart, const BYTE *buf, DWORD dwLength, DWORD *pdwCopied)
{
    DWORD nCopied= 0;
    DWORD addr= dwStart;

    while (nCopied < dwLength)
    {
        DWORD nChunkSize= min(MEM_PAGE_SIZE-(addr&(MEM_PAGE_SIZE-1)), dwLength-nCopied);

        if (IsBadWritePtr((void*)addr, nChunkSize))
            break;
        memcpy((void*)addr, &buf[nCopied], nChunkSize);
        nCopied += nChunkSize;
        addr += nChunkSize;
    }
    *pdwCopied= nCopied;

    return true;
}
bool ReadPhysicalMemory(DWORD dwStart, BYTE *buf, DWORD dwLength, DWORD *pdwCopied)
{
    DWORD nCopied= 0;

    DWORD dwRoundedStart= dwStart&~0xff;
    DWORD dwStartOffset= dwStart-dwRoundedStart;
    DWORD dwRoundedLength= dwLength+dwStartOffset;

    BYTE *pVirt= (BYTE*)VirtualAlloc(NULL, dwRoundedLength, MEM_RESERVE, PAGE_NOACCESS);
    if (pVirt==NULL) {
        *pdwCopied= nCopied;
        return false;
    }

    if (!VirtualCopy(pVirt, (void*)(dwRoundedStart>>8), dwRoundedLength, PAGE_PHYSICAL|PAGE_READONLY|PAGE_NOCACHE)) {
        VirtualFree(pVirt, dwRoundedLength, MEM_DECOMMIT);
        VirtualFree(pVirt, dwRoundedLength, MEM_RELEASE);
        *pdwCopied= nCopied;
        return false;
    }

    DWORD addr= (DWORD)pVirt+dwStartOffset;

    //debug("allocated s=%08lx l=%08lx pv=%08lx a=%08lx\n", dwRoundedStart, dwRoundedLength, pVirt, addr);
    while (nCopied < dwLength)
    {
        DWORD nChunkSize= min(MEM_PAGE_SIZE-(addr&(MEM_PAGE_SIZE-1)), dwLength-nCopied);

        if (IsBadReadPtr((void*)addr, nChunkSize))
            break;
        //debug("copying a=%08lx l=%08lx   done=%08lx\n", addr, nChunkSize, nCopied);
        memcpy(&buf[nCopied], (void*)addr, nChunkSize);
        nCopied += nChunkSize;
        addr += nChunkSize;
    }
    VirtualFree(pVirt, dwRoundedLength, MEM_DECOMMIT);
    VirtualFree(pVirt, dwRoundedLength, MEM_RELEASE);
    *pdwCopied= nCopied;

    return true;
}

ITSUTILS_API HRESULT STDAPICALLTYPE ITReadProcessMemory(
        DWORD cbInput, ReadProcessMemoryParams *pbInput,
        DWORD *pcbOutput, ReadProcessMemoryResult **ppbOutput,
        IRAPIStream *pStream)
{
    ReadProcessMemoryResult *pOut=NULL;

    *pcbOutput= PTR_DIFF(pOut, pOut->buffer+pbInput->nSize);

    pOut= *ppbOutput= (ReadProcessMemoryResult*)LocalAlloc(LPTR, *pcbOutput);

    BOOL bMode = SetKMode(TRUE);
    DWORD dwPerm = SetProcPermissions(0xFFFFFFFF);

    DWORD res= 0;
    if (pbInput->hProcess==INVALID_HANDLE_VALUE)
        ReadPhysicalMemory(pbInput->dwOffset, &pOut->buffer[0], pbInput->nSize, &pOut->dwNumberOfBytesRead);
    else if (pbInput->hProcess==NULL)
        ReadMemory(pbInput->dwOffset, &pOut->buffer[0], pbInput->nSize, &pOut->dwNumberOfBytesRead);
    else if (!ReadProcessMemory(pbInput->hProcess, (void*)pbInput->dwOffset, &pOut->buffer[0], pbInput->nSize, &pOut->dwNumberOfBytesRead))
        res= GetLastError();

    SetProcPermissions(dwPerm);
    SetKMode(bMode);

    return res;
}

ITSUTILS_API HRESULT STDAPICALLTYPE ITWriteProcessMemory(
        DWORD cbInput, WriteProcessMemoryParams *pbInput,
        DWORD *pcbOutput, WriteProcessMemoryResult **ppbOutput,
        IRAPIStream *pStream)
{
    *pcbOutput= sizeof(WriteProcessMemoryResult);
    WriteProcessMemoryResult *pOut= *ppbOutput= (WriteProcessMemoryResult*)LocalAlloc(LPTR, *pcbOutput);

    DWORD res= 0;
    BOOL bMode = SetKMode(TRUE);
    DWORD dwPerm = SetProcPermissions(0xFFFFFFFF);

    if (pbInput->hProcess==NULL)
        WriteMemory(pbInput->dwOffset, &pbInput->buffer[0], pbInput->nSize, &pOut->dwNumberOfBytesWritten);
    else if (!WriteProcessMemory(pbInput->hProcess, (void*)pbInput->dwOffset, &pbInput->buffer[0], pbInput->nSize, &pOut->dwNumberOfBytesWritten))
        res= GetLastError();

    SetProcPermissions(dwPerm);
    SetKMode(bMode);

    return res;
}

ITSUTILS_API HRESULT STDAPICALLTYPE ITGetContext(
        DWORD cbInput, BYTE *pbInput,
        DWORD *pcbOutput, GetContextResult **ppbOutput,
        IRAPIStream *pStream)
{
    WCHAR *cmdline;

//    debug("ITGetContext: i=%08lx(%d)  o=%08lx(%08lx) s=%08lx\n", 
//            cbInput, pbInput, pcbOutput, ppbOutput, pStream);

    cmdline= GetCommandLine();

    GetContextResult *pOut=NULL;

    *pcbOutput= PTR_DIFF(pOut, pOut->wszCmdLine+wcslen(cmdline)+1);

    pOut= *ppbOutput= (GetContextResult*)LocalAlloc(LPTR, *pcbOutput);

    pOut->dwProcessId= GetCurrentProcessId();
    pOut->hProcess= GetCurrentProcess();
    pOut->dwThreadId= GetCurrentThreadId();
    pOut->hThread= GetCurrentThread();
    pOut->dwCallerTrust= CeGetCallerTrust();
    pOut->dwCurrentTrust= CeGetCurrentTrust();
    wcscpy(pOut->wszCmdLine, cmdline);

    return 0;
}

DWORD CalcHeapUsage(HANDLE hTH, DWORD dwProcessID, DWORD dwHeapID)
{
    DWORD dwTotal= 0;
    HEAPENTRY32 he;
    int n=0;

    __try {
        he.dwSize= sizeof(HEAPENTRY32);
        if (Heap32First(hTH, &he, dwProcessID, dwHeapID))
        {
            do {
                dwTotal += he.dwBlockSize;

                he.dwSize= sizeof(HEAPENTRY32);
                n++;
            } while (Heap32Next(hTH, &he));
        }
    }
    __except(1) {
        dwTotal= -1;
    }

    return dwTotal;
}

void GetProcessTimes(DWORD pid, DWORD *ptKernel, DWORD *ptUser)
{
    if (isWince3x())
        GetProcessTimes300(pid, ptKernel, ptUser);
    else
        GetProcessTimes420(pid, ptKernel, ptUser);
}
const WCHAR *GetProcessCommandLine(DWORD pid)
{
    if (isWince3x())
        return GetProcessCommandLine300(pid);
    else
        return GetProcessCommandLine420(pid);
}
void GetProcessUsageList(GetProcessUsageListResult *pOut)
{
    if (isWince3x())
        GetProcessUsageList300(pOut);
    else
        GetProcessUsageList420(pOut);
}
void GetThreadUsageList(GetThreadUsageListResult *pOut)
{
    if (isWince3x())
        GetThreadUsageList300(pOut);
    else
        GetThreadUsageList420(pOut);
}


ITSUTILS_API HRESULT STDAPICALLTYPE ITGetProcessList(
        DWORD cbInput, GetProcessListParams *pbInput,
        DWORD *pcbOutput, GetProcessListResult **ppbOutput,
        IRAPIStream *pStream)
{
    HANDLE hTH= CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS|TH32CS_SNAPHEAPLIST, 0);
    if (hTH==INVALID_HANDLE_VALUE)
        return GetLastError();

    BOOL bMode = SetKMode(TRUE);
    DWORD dwPerm = SetProcPermissions(0xFFFFFFFF);

    GetProcessListResult *pOut=NULL;

    *pcbOutput= PTR_DIFF(pOut,  pOut->pe+MAX_PROCESSES);

    pOut= *ppbOutput= (GetProcessListResult*)LocalAlloc(LPTR, *pcbOutput);

    PROCESSENTRY32 pe;
    pe.dwSize= sizeof(PROCESSENTRY32);

    int procnr=0;
    if (Process32First(hTH, &pe))
    {
        do {
            CEPROCESSENTRY *cepe= &pOut->pe[procnr++];

            cepe->cntUsage= pe.cntUsage; 
            cepe->dwProcessID= pe.th32ProcessID; 
            cepe->dwDefaultHeapID= pe.th32DefaultHeapID; 
            cepe->dwModuleID= pe.th32ModuleID; 
            cepe->cntThreads= pe.cntThreads; 
            cepe->dwParentProcessID= pe.th32ParentProcessID; 
            cepe->dwFlags= pe.dwFlags; 
            cepe->dwMemoryBase= pe.th32MemoryBase;
            cepe->dwAccessKey= pe.th32AccessKey;

            wcsncpy(cepe->szExeFile, pe.szExeFile, MAX_PATH); 
            wcsncpy(cepe->szCmdLine, GetProcessCommandLine(pe.th32ProcessID), MAX_PATH);

            // this is optional, since the the heap of gwes.exe may take upto 5 seconds to scan.
            if (pbInput->bIncludeHeapUsage)
                cepe->dwDefaultHeapSize= CalcHeapUsage(hTH, pe.th32ProcessID, pe.th32DefaultHeapID);

            GetProcessTimes(pe.th32ProcessID, &cepe->dwKernelTime, &cepe->dwUserTime);

        } while (Process32Next(hTH, &pe));
    }

    CloseToolhelp32Snapshot(hTH);

    pOut->nEntries= procnr;

    SetProcPermissions(dwPerm);
    SetKMode(bMode);

    // shrink to desired size.
    *pcbOutput= PTR_DIFF(pOut,  pOut->pe+pOut->nEntries);
    *ppbOutput= (GetProcessListResult*)LocalReAlloc(*ppbOutput, *pcbOutput, LPTR);
    return 0;
}

ITSUTILS_API HRESULT STDAPICALLTYPE ITFindNextMemoryBlock(
        DWORD cbInput, FindNextMemoryBlockParams *pbInput,
        DWORD *pcbOutput, FindNextMemoryBlockResult **ppbOutput,
        IRAPIStream *pStream)
{
    int wrapped=0;

    BOOL bMode = SetKMode(TRUE);
    DWORD dwPerm = SetProcPermissions(0xFFFFFFFF);

    *pcbOutput= sizeof(FindNextMemoryBlockResult);
    FindNextMemoryBlockResult *pOut= *ppbOutput= (FindNextMemoryBlockResult*)LocalAlloc(LPTR, *pcbOutput);

    DWORD addr= pbInput->dwStartSearchAddress;
    while (IsBadReadPtr((void*)addr, 4096))
    {
        if (addr+4096 < addr)
            wrapped= 1;
        addr += 4096;
    }

    pOut->dwBlockStart= addr;
    DWORD len= 0;
    while (!IsBadReadPtr((void*)addr, 4096))
    {
        if (addr+4096 < addr)
            wrapped= 1;
        addr += 4096;
        len += 4096;
    }

    pOut->dwBlockLength= len;

    SetProcPermissions(dwPerm);
    SetKMode(bMode);

    return wrapped;
}


ITSUTILS_API HRESULT STDAPICALLTYPE ITGetProcessHandle(
        DWORD cbInput, GetProcessHandleParams *pbInput,
        DWORD *pcbOutput, GetProcessHandleResult **ppbOutput,
        IRAPIStream *pStream)
{
    *pcbOutput=sizeof(GetProcessHandleResult);
    *ppbOutput=(GetProcessHandleResult*)LocalAlloc(LPTR, *pcbOutput);

    (*ppbOutput)->hProc= GetProcessHandle(pbInput->wszProcessName);

    return 0;
}
HANDLE GetProcessHandle(WCHAR *wszProcessName)
{
    HANDLE hTH= CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);

    PROCESSENTRY32 pe;
    pe.dwSize= sizeof(PROCESSENTRY32);

    HANDLE hProc= INVALID_HANDLE_VALUE;
    if (Process32First(hTH, &pe))
    {
        do {

            if (wcsicmp(wszProcessName, pe.szExeFile)==0)
            {
                hProc= OpenProcess(0, 0, pe.th32ProcessID);
                if (hProc != INVALID_HANDLE_VALUE && hProc!=NULL)
                    break;
            }
        } while (Process32Next(hTH, &pe));
    }

    CloseToolhelp32Snapshot(hTH);

    return hProc;
}


ITSUTILS_API HRESULT STDAPICALLTYPE ITGetThreadTimes(
        DWORD cbInput, GetThreadTimesParams *pbInput,
        DWORD *pcbOutput, GetThreadTimesResult **ppbOutput,
        IRAPIStream *pStream)
{
    *pcbOutput= sizeof(GetThreadTimesResult);
    GetThreadTimesResult *pOut= *ppbOutput= (GetThreadTimesResult *)LocalAlloc(LPTR, *pcbOutput);

    BOOL bMode = SetKMode(TRUE);
    DWORD dwPerm = SetProcPermissions(0xFFFFFFFF);

    DWORD res= 0;
    if (!GetThreadTimes(pbInput->hThread, &pOut->tCreate, &pOut->tExit, &pOut->tKernel, &pOut->tUser))
        res= GetLastError();

    SetProcPermissions(dwPerm);
    SetKMode(bMode);
    return res;
}

ITSUTILS_API HRESULT STDAPICALLTYPE ITGetProcessUsageList(
        DWORD cbInput, GetProcessUsageListParams *pbInput,
        DWORD *pcbOutput, GetProcessUsageListResult **ppbOutput,
        IRAPIStream *pStream)
{
    *pcbOutput= sizeof(GetProcessUsageListResult);
    GetProcessUsageListResult *pOut= *ppbOutput= (GetProcessUsageListResult *)LocalAlloc(LPTR, *pcbOutput);

    memset(pOut, 0, *pcbOutput);

    BOOL bMode = SetKMode(TRUE);
    DWORD dwPerm = SetProcPermissions(0xFFFFFFFF);

    GetProcessUsageList(pOut);

    SetProcPermissions(dwPerm);
    SetKMode(bMode);

    pOut->tQuery= GetTickCount();

    return 0;
}

ITSUTILS_API HRESULT STDAPICALLTYPE ITGetThreadUsageList(
        DWORD cbInput, GetThreadUsageListParams *pbInput,
        DWORD *pcbOutput, GetThreadUsageListResult **ppbOutput,
        IRAPIStream *pStream)
{

    *pcbOutput= sizeof(GetThreadUsageListResult);
    GetThreadUsageListResult *pOut= *ppbOutput= (GetThreadUsageListResult *)LocalAlloc(LPTR, *pcbOutput);

    memset(pOut, 0, *pcbOutput);

    BOOL bMode = SetKMode(TRUE);
    DWORD dwPerm = SetProcPermissions(0xFFFFFFFF);

    GetThreadUsageList(pOut);

    SetProcPermissions(dwPerm);
    SetKMode(bMode);

    pOut->tQuery= GetTickCount();
    return 0;
}


//---------------------debugger----------------------------------
ITSUTILS_API HRESULT STDAPICALLTYPE ITStartDebuggingProcess(
        DWORD cbInput, StartDebuggingProcessParams *pbInput,
        DWORD *pcbOutput, StartDebuggingProcessResult **ppbOutput,
        IRAPIStream *pStream)
{
    DWORD res= 0;
    *pcbOutput= 0;
    *ppbOutput= NULL;

    HMODULE hDll= LoadLibrary(L"ItsUtils.dll");
    if (hDll==NULL)
        return GetLastError();
    LPTHREAD_START_ROUTINE pProc= (LPTHREAD_START_ROUTINE)GetProcAddress(hDll, L"DebugThreadProc");
    if (pProc==NULL)
    {
        res= GetLastError();
        if (!FreeLibrary(hDll))
            error("ITStartDebuggingProcess: proc not found: FreeLibrary(hDll)");
        //debug("freed library: %08lx\n", hDll);
        return res;
    }
    //debug("loaded library: h=%08lx proc=%08lx\n", hDll, pProc);

    DebuggerInfo *di= (DebuggerInfo *)LocalAlloc(LPTR, sizeof(DebuggerInfo));
    memcpy(di->magic, "ITDB", 4);
    di->evStarted= CreateEvent(NULL, false, false, NULL);   // used only once
    di->evBuffer= INVALID_HANDLE_VALUE;                     // signal new buffer entry
    di->smBuffer= INVALID_HANDLE_VALUE;                     // locks buffer info
    di->evStopped= INVALID_HANDLE_VALUE;                    // used only once
    di->dwThreadResult= 0;
    di->dwDebuggedProcessId= pbInput->dwProcessId;

    wcsncpy(di->wszExeName, pbInput->wszExeName, MAX_PATH);
    wcsncpy(di->wszCmdLine, pbInput->wszCmdLine, MAX_PATH);

    di->dwMissedEvents= 0;

    di->hDll= hDll;

    di->bufsize= 0;
    di->bufhead= 0;
    di->buftail= 0;
    di->bufmaxsize= MAX_DEBUG_EVENTS;

    di->bThreadStarted= false;
    di->bThreadTerminated= false;
    di->hThread= CreateThread(NULL, 0, pProc, di, 0, &di->dwThreadId);

    //debug("created debug thread: h=%08lx id=%08lx\n", di->hThread, di->dwThreadId);

    // wait until debug thread is active
    bool bOk= (WaitForSingleObject(di->evStarted, INFINITE)==WAIT_OBJECT_0);
    CloseHandle(di->evStarted);
    di->evStarted= INVALID_HANDLE_VALUE;

    *pcbOutput= sizeof(StartDebuggingProcessResult);
    StartDebuggingProcessResult *pOut= *ppbOutput= (StartDebuggingProcessResult *)LocalAlloc(LPTR, *pcbOutput);

    pOut->debugger= (DWORD)di;

    if (di->dwThreadResult)
        return di->dwThreadResult;
    else if (!di->bThreadStarted)
    {
        debug("error starting debug thread\n");
        return -1;
    }
                                    
    return 0;
}

ITSUTILS_API HRESULT STDAPICALLTYPE ITStopDebuggingProcess(
        DWORD cbInput, StopDebuggingProcessParams *pbInput,
        DWORD *pcbOutput, StopDebuggingProcessResult **ppbOutput,
        IRAPIStream *pStream)
{
    DebuggerInfo *di= (DebuggerInfo *)pbInput->debugger;

    *pcbOutput= 0;
    *ppbOutput= NULL;

    //debug("ITStopDebuggingProcess\n");
    if (IsBadReadPtr(di, sizeof(DebuggerInfo)))
    {
        debug("ITStopDebuggingProcess: di is invalid ptr: %08lx\n", di);
        return -1;
    }
    if (memcmp(di->magic, "ITDB", 4))
    {
        debug("ITStopDebuggingProcess: di struct is not valid: %02x%02x%02x%02x\n", di->magic[0], di->magic[1], di->magic[2], di->magic[3]);
        return -1;
    }

    if (!di->bThreadTerminated)
    {
        //debug("ITStopDebuggingProcess: stopping debug thread\n");
        // creation of event signals to thread that it must stop.
        di->evStopped= CreateEvent(NULL, true, false, NULL);     // automatic event

        // wait 5 sec until thread really stopped, before freeing up memory.
        bool bOk= (WaitForSingleObject(di->evStopped, 5000)==WAIT_OBJECT_0);
        CloseHandle(di->evStopped);
        di->evStopped= INVALID_HANDLE_VALUE;

        DWORD nExitCode;
        GetExitCodeThread(di->hThread, &nExitCode);
        TerminateThread(di->hThread, nExitCode);
    }

    if (!CloseHandle(di->hThread))
        error("stopdebug: CloseHandle(hThread)");

    if (!FreeLibrary(di->hDll))
        error("stopdebug: FreeLibrary(hDll)");

    //debug("freed library: %08lx\n", di->hDll);

    memset(di->magic, 0, 4);
    LocalFree(di);

    return 0;
}


ITSUTILS_API HRESULT STDAPICALLTYPE ITWaitForDebugEvent(
        DWORD cbInput, WaitForDebugEventParams *pbInput,
        DWORD *pcbOutput, WaitForDebugEventResult **ppbOutput,
        IRAPIStream *pStream)
{
    DWORD res=0;
    DebuggerInfo *di= (DebuggerInfo *)pbInput->debugger;

    *pcbOutput= 0;
    *ppbOutput= NULL;


    if (IsBadReadPtr(di, sizeof(DebuggerInfo)))
    {
        debug("ITWaitForDebugEvent : di is invalid ptr: %08lx\n", di);
        return -1;
    }
    if (memcmp(di->magic, "ITDB", 4))
    {
        debug("ITWaitForDebugEvent : di struct is not valid: %02x%02x%02x%02x\n", di->magic[0], di->magic[1], di->magic[2], di->magic[3]);
        return -1;
    }

    if (!di->bThreadStarted)
    {
        debug("ITWaitForDebugEvent : debug thread not started\n");
        return -1;
    }
    if (di->bThreadTerminated)
    {
        //debug("ITWaitForDebugEvent : debug thread terminated\n");
        return -1;
    }

    *pcbOutput= sizeof(WaitForDebugEventResult);
    WaitForDebugEventResult *pOut= *ppbOutput= (WaitForDebugEventResult *)LocalAlloc(LPTR, *pcbOutput);

    // either there is already something in the buffer, otherwise wait for event.
    if (!di->bufsize==0 
            || (WaitForSingleObject(di->evBuffer, INFINITE)==WAIT_OBJECT_0
                && di->bThreadStarted && !di->bThreadTerminated))
    {
        // wait until we are allowed to touch the buffer internals.
        if (WaitForSingleObject(di->smBuffer, INFINITE)==WAIT_OBJECT_0)
        {
            if (!GetDebugEvent(di, &pOut->ev)) {
                debug("ITWaitForDebugEvent : no debug event found\n");
                res= -1;
            }

            if (!ReleaseSemaphore(di->smBuffer, 1, NULL))
            {
                res= GetLastError();
                error(res, "ITWaitForDebugEvent : ReleaseSemaphore(smbuffer)");
            }
        }
        if (!ResetEvent(di->evBuffer))
        {
            res= GetLastError();
            error(res, "ITWaitForDebugEvent : ResetEvent(evBuffer)");
        }
    }

    pOut->dwMissedEvents= di->dwMissedEvents;

    return res;
}


ITSUTILS_API DWORD DebugThreadProc(DebuggerInfo *di)
{
    DWORD res=0;
    //BOOL bMode = SetKMode(TRUE);
    DWORD dwPerm = SetProcPermissions(0xFFFFFFFF);

    //debug("DebugThreadProc started\n");

    // either use provided pid, or start new process
    if (di->dwDebuggedProcessId!=0)
    {
        if (!DebugActiveProcess(di->dwDebuggedProcessId))
        {
            res= di->dwThreadResult= GetLastError();
        }
    }
    else
    {
        PROCESS_INFORMATION pi;
        if (!CreateProcess(di->wszExeName, di->wszCmdLine, NULL, NULL, 0, DEBUG_PROCESS,
                NULL, NULL, NULL, &pi))
        {
            res= di->dwThreadResult= GetLastError();
        }
        else {
            di->dwDebuggedProcessId= pi.dwProcessId;
        }
    }

    di->bProcessTerminated= false;

    di->evBuffer= CreateEvent(NULL, false, false, NULL);     // signal new buffer entry
    di->smBuffer= CreateSemaphore(NULL, 1, 1, NULL);         // locks buffer info

    // tell parent process that we are initialized and ready
    if (di->evStarted!=INVALID_HANDLE_VALUE && !SetEvent(di->evStarted))
    {
        res= di->dwThreadResult= GetLastError();
    }

    di->bThreadStarted= true;

    DEBUG_EVENT ev;

    // loop until a 'stopped' event is created
    while(!res && di->evStopped==INVALID_HANDLE_VALUE)
    {
        if (WaitForDebugEvent(&ev, 1000))
        {
            if (!ProcessDebugEvent(di, &ev))
                di->dwMissedEvents++;

            if (di->bProcessTerminated)
                break;
            if (!ContinueDebugEvent(ev.dwProcessId, ev.dwThreadId, DBG_EXCEPTION_NOT_HANDLED ))
                res= di->dwThreadResult= GetLastError();
        }
    }
    di->bThreadTerminated= true;
    //debug("DebugThreadProc terminated\n");

    // 'stopped' event found, set it.
    if (di->evStopped!=INVALID_HANDLE_VALUE && !SetEvent(di->evStopped))
    {
        res= GetLastError();
        error(res, "SetEvent(evStopped)");
    }

    // try to tell whoever is waiting for us to stop doing that.
    if (di->evBuffer!=INVALID_HANDLE_VALUE)
    {
        // wake up other thread to be able to see me.
        SetEvent(di->evBuffer);
        CloseHandle(di->evBuffer);
        di->evBuffer= INVALID_HANDLE_VALUE;
    }
    if (di->smBuffer!=INVALID_HANDLE_VALUE)
    {
        CloseHandle(di->smBuffer);
        di->smBuffer= INVALID_HANDLE_VALUE;
    }

    SetProcPermissions(dwPerm);
    //SetKMode(bMode);

    return res;
}

bool GetDebugEvent(DebuggerInfo *di, DebugEvent *ev)
{
    if (di->bufsize==0)
        return false;

    memcpy(ev, &di->buffer[di->bufhead], sizeof(DebugEvent));

    if (++di->bufhead==di->bufmaxsize)
        di->bufhead= 0;

    di->bufsize--;

    return true;
}

bool ProcessDebugEvent(DebuggerInfo *di, DEBUG_EVENT *sysev)
{
    if (di->bufsize==di->bufmaxsize)
        return false;
    if (WaitForSingleObject(di->smBuffer, INFINITE)==WAIT_OBJECT_0)
    {
        if (di->bufsize==di->bufmaxsize)
            return false;

        DebugEvent *ev= &di->buffer[di->buftail];

        ev->dwProcessId= sysev->dwProcessId;
        ev->dwThreadId= sysev->dwThreadId;
        ev->dwDebugEventCode= sysev->dwDebugEventCode;

        ParseDebugEvent(ev->description, sizeof(ev->description), sysev);

        if (++di->buftail==di->bufmaxsize)
            di->buftail= 0;

        ++di->bufsize;

        // todo: test sysev.ExceptionRecord.ExceptionFlags&EXCEPTION_NONCONTINUABLE 
        if (ev->dwDebugEventCode==EXIT_PROCESS_DEBUG_EVENT)
            di->bProcessTerminated= true;

        if (!ReleaseSemaphore(di->smBuffer, 1, NULL))
            return false;
        if (!SetEvent(di->evBuffer))
            return false;
    }

    return true;
}

//--------------converting a debug event to a string
int ParseDebugException(char *buf, int bufsize, DWORD dwProcessId, DWORD dwThreadId, EXCEPTION_DEBUG_INFO *exception);
int ParseDebugCreateThread(char *buf, int bufsize, DWORD dwProcessId, DWORD dwThreadId, CREATE_THREAD_DEBUG_INFO *createThread);
int ParseDebugCreateProcessInfo(char *buf, int bufsize, DWORD dwProcessId, DWORD dwThreadId, CREATE_PROCESS_DEBUG_INFO *createProcessInfo);
int ParseDebugExitThread(char *buf, int bufsize, DWORD dwProcessId, DWORD dwThreadId, EXIT_THREAD_DEBUG_INFO *exitThread);
int ParseDebugExitProcess(char *buf, int bufsize, DWORD dwProcessId, DWORD dwThreadId, EXIT_PROCESS_DEBUG_INFO *exitProcess);
int ParseDebugLoadDll(char *buf, int bufsize, DWORD dwProcessId, DWORD dwThreadId, LOAD_DLL_DEBUG_INFO *loadDll);
int ParseDebugUnloadDll(char *buf, int bufsize, DWORD dwProcessId, DWORD dwThreadId, UNLOAD_DLL_DEBUG_INFO *unloadDll);
int ParseDebugString(char *buf, int bufsize, DWORD dwProcessId, DWORD dwThreadId, OUTPUT_DEBUG_STRING_INFO *debugString);
int ParseDebugRipInfo(char *buf, int bufsize, DWORD dwProcessId, DWORD dwThreadId, RIP_INFO *ripInfo);


int ParseDebugEvent(char *buf, int bufsize, DEBUG_EVENT *event)
{
    int n=0;
    switch(event->dwDebugEventCode)
    {
    case EXCEPTION_DEBUG_EVENT:     n+=ParseDebugException(buf+n, bufsize-n, event->dwProcessId, event->dwThreadId, &event->u.Exception); break;
    case CREATE_THREAD_DEBUG_EVENT: n+=ParseDebugCreateThread(buf+n, bufsize-n, event->dwProcessId, event->dwThreadId, &event->u.CreateThread); break;
    case CREATE_PROCESS_DEBUG_EVENT:n+=ParseDebugCreateProcessInfo(buf+n, bufsize-n, event->dwProcessId, event->dwThreadId, &event->u.CreateProcessInfo); break;
    case EXIT_THREAD_DEBUG_EVENT:   n+=ParseDebugExitThread(buf+n, bufsize-n, event->dwProcessId, event->dwThreadId, &event->u.ExitThread); break;
    case EXIT_PROCESS_DEBUG_EVENT:  n+=ParseDebugExitProcess(buf+n, bufsize-n, event->dwProcessId, event->dwThreadId, &event->u.ExitProcess); break;
    case LOAD_DLL_DEBUG_EVENT:      n+=ParseDebugLoadDll(buf+n, bufsize-n, event->dwProcessId, event->dwThreadId, &event->u.LoadDll); break;
    case UNLOAD_DLL_DEBUG_EVENT:    n+=ParseDebugUnloadDll(buf+n, bufsize-n, event->dwProcessId, event->dwThreadId, &event->u.UnloadDll); break;
    case OUTPUT_DEBUG_STRING_EVENT: n+=ParseDebugString(buf+n, bufsize-n, event->dwProcessId, event->dwThreadId, &event->u.DebugString); break;
    case RIP_EVENT:                 n+=ParseDebugRipInfo(buf+n, bufsize-n, event->dwProcessId, event->dwThreadId, &event->u.RipInfo); break;
    default:
        n+= _snprintf(buf+n, bufsize-n, "unknown debug event pid=%08lx tid=%08lx ev=%08lx", event->dwProcessId, event->dwThreadId, event->dwDebugEventCode);
    }
    return n;
}

// provide some info in case other things fail.
TCHAR *StrInfo(DWORD dwProcessId, void *ptr, int len, int uni)
{
    static TCHAR tbuf[256];
    _sntprintf(tbuf, 256, L"[%08lx:%08lx L%04x uni=%d]", dwProcessId, ptr, len, uni);
    return tbuf;
}
TCHAR *ReadProcessWString(DWORD dwProcessId, void *ptr, int len)
{
    static TCHAR buf[1024];
    DWORD nCopied;
    if (!ReadProcessMemory((HANDLE)dwProcessId, ptr, buf, min(len, sizeof(buf)-1), &nCopied))
        return 0;
    buf[nCopied]= 0;

    return buf;
}
char *ReadProcessString(DWORD dwProcessId, void *ptr, int len)
{
    static char buf[1024];
    DWORD nCopied;
    if (!ReadProcessMemory((HANDLE)dwProcessId, ptr, buf, min(len, sizeof(buf)-1), &nCopied))
        return 0;
    buf[nCopied]= 0;

    return buf;
}
DWORD ReadProcessDword(DWORD dwProcessId, void *ptr)
{
    DWORD dw;
    DWORD nCopied;
    if (!ReadProcessMemory((HANDLE)dwProcessId, ptr, &dw, sizeof(DWORD), &nCopied))
        return 0;

    return dw;
}
void FindModuleNameForProcessAddress(DWORD pid, DWORD dwAddress, TCHAR *szModname, int maxmodname)
{
	HANDLE hTH= CreateToolhelp32Snapshot(TH32CS_SNAPMODULE, pid);

	MODULEENTRY32 me;
	me.dwSize= sizeof(MODULEENTRY32);

    szModname[0]= 0;

	if (Module32First(hTH, &me))
	{
		do {
			if ((DWORD)me.modBaseAddr<=dwAddress && dwAddress<(DWORD)me.modBaseAddr+me.modBaseSize)
            {
                if (me.szExePath[0])
                    _snwprintf(szModname, maxmodname, L"%ls\\%ls", me.szExePath, me.szModule);
                else
                    wcsncpy(szModname, me.szModule, maxmodname);
                break;
            }

			me.dwSize= sizeof(MODULEENTRY32);
		} while (Module32Next(hTH, &me));
	}
    CloseToolhelp32Snapshot(hTH);
}

TCHAR *PtrToString(DWORD dwProcessId, void *ptr, int len, int uni)
{
    if (ptr==NULL)
        return L"(NULL)";
    else if (uni)
    {
        TCHAR *tbuf= ReadProcessWString(dwProcessId, ptr, len);
        if (tbuf==NULL)
            return StrInfo(dwProcessId, ptr, len, uni);

        return tbuf;
    }
    else {
        static TCHAR tbuf[256];
        char *buf= ReadProcessString(dwProcessId, ptr, len);

        if (buf==NULL)
            return StrInfo(dwProcessId, ptr, len, uni);
        _sntprintf(tbuf, 256, L"%hs", buf);
        return tbuf;
    }
}
TCHAR *PtrptrToString(DWORD dwProcessId, void *pptr, int len, int uni)
{
    if (pptr==NULL)
        return L"((NULL))";

    void *ptr= (void*)ReadProcessDword(dwProcessId, pptr);

    return PtrToString(dwProcessId, ptr, len, uni);
}
int ParseExceptionCode(char *buf, int bufsize, DWORD dwCode)
{
    const char *str=NULL;
    int n=0;
    switch (dwCode)
    {
    case EXCEPTION_ACCESS_VIOLATION:            str= "EXCEPTION_ACCESS_VIOLATION"; break;
    case EXCEPTION_DATATYPE_MISALIGNMENT:       str= "EXCEPTION_DATATYPE_MISALIGNMENT"; break;
    case EXCEPTION_BREAKPOINT:                  str= "EXCEPTION_BREAKPOINT"; break;
    case EXCEPTION_SINGLE_STEP:                 str= "EXCEPTION_SINGLE_STEP"; break;
    case EXCEPTION_ARRAY_BOUNDS_EXCEEDED:       str= "EXCEPTION_ARRAY_BOUNDS_EXCEEDED"; break;
    case EXCEPTION_FLT_DENORMAL_OPERAND:        str= "EXCEPTION_FLT_DENORMAL_OPERAND"; break;
    case EXCEPTION_FLT_DIVIDE_BY_ZERO:          str= "EXCEPTION_FLT_DIVIDE_BY_ZERO"; break;
    case EXCEPTION_FLT_INEXACT_RESULT:          str= "EXCEPTION_FLT_INEXACT_RESULT"; break;
    case EXCEPTION_FLT_INVALID_OPERATION:       str= "EXCEPTION_FLT_INVALID_OPERATION"; break;
    case EXCEPTION_FLT_OVERFLOW:                str= "EXCEPTION_FLT_OVERFLOW"; break;
    case EXCEPTION_FLT_STACK_CHECK:             str= "EXCEPTION_FLT_STACK_CHECK"; break;
    case EXCEPTION_FLT_UNDERFLOW:               str= "EXCEPTION_FLT_UNDERFLOW"; break;
    case EXCEPTION_INT_DIVIDE_BY_ZERO:          str= "EXCEPTION_INT_DIVIDE_BY_ZERO"; break;
    case EXCEPTION_INT_OVERFLOW:                str= "EXCEPTION_INT_OVERFLOW"; break;
    case EXCEPTION_PRIV_INSTRUCTION:            str= "EXCEPTION_PRIV_INSTRUCTION"; break;
    case EXCEPTION_IN_PAGE_ERROR:               str= "EXCEPTION_IN_PAGE_ERROR"; break;
    case EXCEPTION_ILLEGAL_INSTRUCTION:         str= "EXCEPTION_ILLEGAL_INSTRUCTION"; break;
    case EXCEPTION_NONCONTINUABLE_EXCEPTION:    str= "EXCEPTION_NONCONTINUABLE_EXCEPTION"; break;
    case EXCEPTION_STACK_OVERFLOW:              str= "EXCEPTION_STACK_OVERFLOW"; break;
    case EXCEPTION_INVALID_DISPOSITION:         str= "EXCEPTION_INVALID_DISPOSITION"; break;
    case EXCEPTION_GUARD_PAGE:                  str= "EXCEPTION_GUARD_PAGE"; break;
    case EXCEPTION_INVALID_HANDLE:              str= "EXCEPTION_INVALID_HANDLE"; break;
    }
    if (str)
        n+= _snprintf(buf+n, bufsize-n, " %s", str);
    else
        n+= _snprintf(buf+n, bufsize-n, " EXCEPTION_%08lx", dwCode);

    return n;
}
int ParseExceptionFlags(char *buf, int bufsize, DWORD dwFlags)
{
    int n=0;
    if (dwFlags==0) {
        n+= _snprintf(buf+n, bufsize-n, " noflags");
        return n;
    }
    n+= _snprintf(buf+n, bufsize-n, " flags=");
    int m=n;
    if (dwFlags&EXCEPTION_NONCONTINUABLE ) n+= _snprintf(buf+n, bufsize-n, "%sNONCONTINUABLE", n==m?"|":"");
    if (dwFlags&EXCEPTION_UNWINDING      ) n+= _snprintf(buf+n, bufsize-n, "%sUNWINDING", n==m?"|":"");
    if (dwFlags&EXCEPTION_EXIT_UNWIND    ) n+= _snprintf(buf+n, bufsize-n, "%sEXIT_UNWIND", n==m?"|":"");
    if (dwFlags&EXCEPTION_STACK_INVALID  ) n+= _snprintf(buf+n, bufsize-n, "%sSTACK_INVALID", n==m?"|":"");
    if (dwFlags&EXCEPTION_NESTED_CALL    ) n+= _snprintf(buf+n, bufsize-n, "%sNESTED_CALL", n==m?"|":"");
    if (dwFlags&EXCEPTION_TARGET_UNWIND  ) n+= _snprintf(buf+n, bufsize-n, "%sTARGET_UNWIND", n==m?"|":"");
    if (dwFlags&EXCEPTION_COLLIDED_UNWIND) n+= _snprintf(buf+n, bufsize-n, "%sCOLLIDED_UNWIND", n==m?"|":"");

    const DWORD knownflags= EXCEPTION_NONCONTINUABLE|EXCEPTION_UNWINDING|EXCEPTION_EXIT_UNWIND|EXCEPTION_STACK_INVALID|EXCEPTION_NESTED_CALL|EXCEPTION_TARGET_UNWIND|EXCEPTION_COLLIDED_UNWIND;
    if (dwFlags&~knownflags) n+= _snprintf(buf+n, bufsize-n, "%s%08lx", n==m?"|":"", dwFlags&~knownflags);

    return n;
}
int ParseDebugException(char *buf, int bufsize, DWORD dwProcessId, DWORD dwThreadId, EXCEPTION_DEBUG_INFO *exception)
{
    int n=0;
    n+= _snprintf(buf+n, bufsize-n, "debugException first=%d", exception->dwFirstChance);

    EXCEPTION_RECORD *er= &exception->ExceptionRecord;
    int i=0;
    while (er)
    {
        if (i)
            n+= _snprintf(buf+n, bufsize-n, " ; ");

        n+= ParseExceptionCode(buf+n, bufsize-n, er->ExceptionCode);
        n+= ParseExceptionFlags(buf+n, bufsize-n, er->ExceptionFlags);
        n+= _snprintf(buf+n, bufsize-n, " @%08lx", er->ExceptionAddress);
            
        if (er->ExceptionCode==EXCEPTION_ACCESS_VIOLATION && er->NumberParameters==3) {
            n+= _snprintf(buf+n, bufsize-n, " IP=%08lx ref=%08lx  desc=%08lx\n", 
                    er->ExceptionInformation[0],
                    er->ExceptionInformation[1],
                    er->ExceptionInformation[2]);
        }
        else if (er->ExceptionCode==EXCEPTION_IN_PAGE_ERROR && er->NumberParameters==3) {
            n+= _snprintf(buf+n, bufsize-n, " IP=%08lx ref=%08lx  status=%08lx\n", 
                    er->ExceptionInformation[0],
                    er->ExceptionInformation[1],
                    er->ExceptionInformation[2]);
        }
        else {
            for(DWORD i=0 ; i<er->NumberParameters ; i++)
            {
                n+= _snprintf(buf+n, bufsize-n, " %08lx", er->ExceptionInformation[i]);
            }
        }
        er= er->ExceptionRecord;
        i++;
    }
    return n;
}

int ParseDebugCreateThread(char *buf, int bufsize, DWORD dwProcessId, DWORD dwThreadId, CREATE_THREAD_DEBUG_INFO *createThread)
{
    int n=0;
    n+= _snprintf(buf+n, bufsize-n, "createThread(hthread=%08x base=%08x start=%08x)", 
        createThread->hThread,
        createThread->lpThreadLocalBase,
        createThread->lpStartAddress);
    return n;
}

int ParseDebugCreateProcessInfo(char *buf, int bufsize, DWORD dwProcessId, DWORD dwThreadId, CREATE_PROCESS_DEBUG_INFO *createProcessInfo)
{
    int n=0;
    n+= _snprintf(buf+n, bufsize-n, "createProcess(file=%d hproc=%08x hthread=%08x image=%08x debug=%08x(%08x)", 
            createProcessInfo->hFile,
            createProcessInfo->hProcess,
            createProcessInfo->hThread,
            createProcessInfo->lpBaseOfImage,
            createProcessInfo->dwDebugInfoFileOffset,
            createProcessInfo->nDebugInfoSize);
    n+= _snprintf(buf+n, bufsize-n, "       base=%08x start=%08x name='%ls'", 
            createProcessInfo->lpThreadLocalBase,
            createProcessInfo->lpStartAddress,
            PtrToString(
                    dwProcessId,
                    createProcessInfo->lpImageName,
                    128,        // todo ..... what should this be ??
                    createProcessInfo->fUnicode));
    return n;
}

int ParseDebugExitThread(char *buf, int bufsize, DWORD dwProcessId, DWORD dwThreadId, EXIT_THREAD_DEBUG_INFO *exitThread)
{
    int n=0;
    // todo: remember 
    n+= _snprintf(buf+n, bufsize-n, "exitThread code=%d", exitThread->dwExitCode);
    return n;
}

int ParseDebugExitProcess(char *buf, int bufsize, DWORD dwProcessId, DWORD dwThreadId, EXIT_PROCESS_DEBUG_INFO *exitProcess)
{
    int n=0;
    n+= _snprintf(buf+n, bufsize-n, "exitProcess code=%d", exitProcess->dwExitCode);
    return n;
}

int ParseDebugLoadDll(char *buf, int bufsize, DWORD dwProcessId, DWORD dwThreadId, LOAD_DLL_DEBUG_INFO *loadDll)
{
    int n=0;
    n+= _snprintf(buf+n, bufsize-n, "loadDll file=%d base=%08x debug=%08x(%08x) name='%ls'", 
            loadDll->hFile,
            loadDll->lpBaseOfDll,
            loadDll->dwDebugInfoFileOffset,
            loadDll->nDebugInfoSize,
            PtrToString(
                    dwProcessId,
                    loadDll->lpImageName,
                    128,        // todo ..... what should this be ??
                    loadDll->fUnicode));
    return n;
}

int ParseDebugUnloadDll(char *buf, int bufsize, DWORD dwProcessId, DWORD dwThreadId, UNLOAD_DLL_DEBUG_INFO *unloadDll)
{
    int n=0;
    // todo: keep track of pid+dllbase to remember the module name.
    n+= _snprintf(buf+n, bufsize-n, "unloadDll base=%08x", unloadDll->lpBaseOfDll);
    return n;
}

int ParseDebugString(char *buf, int bufsize, DWORD dwProcessId, DWORD dwThreadId, OUTPUT_DEBUG_STRING_INFO *debugString)
{
    int n=0;
    n+= _snprintf(buf+n, bufsize-n, "debugString '%ls'", 
            PtrToString(dwProcessId,
                    debugString->lpDebugStringData,
                    debugString->nDebugStringLength,
                    debugString->fUnicode));
    return n;
}

int ParseDebugRipInfo(char *buf, int bufsize, DWORD dwProcessId, DWORD dwThreadId, RIP_INFO *ripInfo)
{
    int n=0;
    n+= _snprintf(buf+n, bufsize-n, "ripinfo error=%d type=%d", ripInfo->dwError, ripInfo->dwType);
    return n;
}

//--------------------------------
int GetDiskInfo(HANDLE hDsk, DISK_INFO &info)
{
    DWORD nReturned;

    // various ce versions use all possible combinations, so we try 'm all.
    if (DeviceIoControl(hDsk, IOCTL_DISK_GETINFO, &info, sizeof(info), NULL, 0, &nReturned, NULL))
        return 1;
    if (DeviceIoControl(hDsk, IOCTL_DISK_GETINFO, NULL, 0, &info, sizeof(info), &nReturned, NULL))
        return 2;
    if (DeviceIoControl(hDsk, DISK_IOCTL_GETINFO, &info, sizeof(info), NULL, 0, &nReturned, NULL))
        return 3;
    if (DeviceIoControl(hDsk, DISK_IOCTL_GETINFO, NULL, 0, &info, sizeof(info), &nReturned, NULL))
        return 4;

    return 0;
}
HANDLE OpenDisk(int n)
{
    TCHAR diskname[16];

    if (n>=0 && n<=256)
        _sntprintf(diskname, 16, L"DSK%u:", n);
    else
        _sntprintf(diskname, 16, L"DSK:");

    HANDLE hDsk= CreateFile( diskname,
            GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,
            NULL, OPEN_EXISTING, 0, 0 );
    if (hDsk==NULL || hDsk==INVALID_HANDLE_VALUE)
    {
        return NULL;
    }
    return hDsk;
}

/*
    TCHAR ss[256]; _sntprintf(ss, 256, L"req: %d %08lx %08lx", pbInput->dwDiskNr, pbInput->dwOffset, pbInput->dwSize);
    MessageBox(0, ss, L"req", 0);
    _sntprintf(ss, 256, L"reply: %d %08lx %08lx -> %d, %d, %08lx %08lx", 
            pbInput->dwDiskNr, pbInput->dwOffset, pbInput->dwSize,
            out->dwNumberOfBytesRead, out->start, out->number, outsize);
    MessageBox(0, ss, L"reply", 0);
 */
bool DiskReadBlocks(HANDLE hDsk, int nStart, int nBlocks, const DISK_INFO& info,
                BYTE* buffer, DWORD* pnBytesRead)
{
    SG_REQ req;

    req.sr_start= nStart;
    req.sr_num_sec= nBlocks;
    req.sr_num_sg= 1;
    req.sr_status= 0;
    req.sr_callback= NULL;
    req.sr_sglist[0].sb_len= nBlocks*info.di_bytes_per_sect;
    req.sr_sglist[0].sb_buf= buffer;

    DWORD nReturned=0;
    if (!DeviceIoControl(hDsk, DISK_IOCTL_READ, &req, sizeof(req), NULL, NULL, &nReturned, NULL))
        return false;

    *pnBytesRead= nBlocks*info.di_bytes_per_sect;

    return true;
}
bool DiskWriteBlocks(HANDLE hDsk, int nStart, int nBlocks, const DISK_INFO& info,
            const BYTE* buffer, DWORD* pnBytesWritten)
{
    SG_REQ req;

    req.sr_start= nStart;
    req.sr_num_sec= nBlocks;
    req.sr_num_sg= 1;
    req.sr_status= 0;
    req.sr_callback= NULL;
    req.sr_sglist[0].sb_len= nBlocks*info.di_bytes_per_sect;
    req.sr_sglist[0].sb_buf= (BYTE*)buffer;

    DWORD nReturned=0;
    if (!DeviceIoControl(hDsk, DISK_IOCTL_WRITE, &req, sizeof(req), NULL, NULL, &nReturned, NULL))
        return false;

    *pnBytesWritten= nBlocks*info.di_bytes_per_sect;

    return true;
}


ITSUTILS_API HRESULT STDAPICALLTYPE ITSDCardInfo(
        DWORD cbInput, SDCardInfoParams *pbInput,
        DWORD *pcbOutput, SDCardInfoResult **ppbOutput,
        IRAPIStream *pStream)
{
    HRESULT res= 0;

    *ppbOutput= NULL;
    *pcbOutput= 0;

    HANDLE hDsk= OpenDisk(pbInput->dwDiskNr);
    if (hDsk==NULL)
        return GetLastError();
    DISK_INFO info;
    if (!GetDiskInfo(hDsk, info))
    {
        res= GetLastError();
        CloseHandle(hDsk);
        return res;
    }
    DWORD outsize= sizeof(SDCardInfoResult);
    SDCardInfoResult *out= (SDCardInfoResult*)LocalAlloc(LPTR, outsize);

    out->blockSize= info.di_bytes_per_sect;
    out->totalBlocks= info.di_total_sectors;

    CloseHandle(hDsk);
    *ppbOutput= out;
    *pcbOutput= outsize;

    return 0;
}

ITSUTILS_API HRESULT STDAPICALLTYPE ITReadSDCard(
        DWORD cbInput, ReadSDCardParams *pbInput,
        DWORD *pcbOutput, ReadSDCardResult **ppbOutput,
        IRAPIStream *pStream)
{
    HRESULT res= 0;

    *ppbOutput= NULL;
    *pcbOutput= 0;

    HANDLE hDsk= OpenDisk(pbInput->dwDiskNr);
    if (hDsk==NULL)
        return GetLastError();
    DISK_INFO info;
    if (!GetDiskInfo(hDsk, info))
    {
        res= GetLastError();
        CloseHandle(hDsk);
        return res;
    }
    if (pbInput->dwOffset % info.di_bytes_per_sect) {
        debug("ERROR: ITReadSDCard: startoffset not on sector boundary: sector nr %x ofs %04x\n",
            pbInput->dwOffset / info.di_bytes_per_sect, pbInput->dwOffset % info.di_bytes_per_sect);
        CloseHandle(hDsk);
        return ERROR_INVALID_PARAMETER;
    }
    if ((pbInput->dwOffset+pbInput->dwSize) % info.di_bytes_per_sect) {
        debug("ERROR: ITReadSDCard: endoffset not on sector boundary: sector nr %x ofs %04x\n",
            (pbInput->dwOffset+pbInput->dwSize)/ info.di_bytes_per_sect, (pbInput->dwOffset+pbInput->dwSize)% info.di_bytes_per_sect);
        CloseHandle(hDsk);
        return ERROR_INVALID_PARAMETER;
    }


    DWORD outsize= sizeof(ReadSDCardResult)+pbInput->dwSize;
    ReadSDCardResult *out= (ReadSDCardResult *)LocalAlloc(LPTR, outsize);

    if (!DiskReadBlocks(hDsk, pbInput->dwOffset / info.di_bytes_per_sect, pbInput->dwSize / info.di_bytes_per_sect, info, out->buffer, &out->dwNumberOfBytesRead))
    {
        res= GetLastError();
        CloseHandle(hDsk);
        LocalFree(out);
        return res;
    }

//    out->start= pbInput->dwOffset / info.di_bytes_per_sect;
//    out->number= pbInput->dwSize / info.di_bytes_per_sect;

    CloseHandle(hDsk);
    *ppbOutput= out;
    *pcbOutput= outsize;

    return 0;
}


ITSUTILS_API HRESULT STDAPICALLTYPE ITWriteSDCard(
        DWORD cbInput, WriteSDCardParams *pbInput,
        DWORD *pcbOutput, WriteSDCardResult **ppbOutput,
        IRAPIStream *pStream)
{
    HRESULT res= 0;

    *ppbOutput= NULL;
    *pcbOutput= 0;

    HANDLE hDsk= OpenDisk(pbInput->dwDiskNr);
    if (hDsk==NULL)
        return GetLastError();
    DISK_INFO info;
    if (!GetDiskInfo(hDsk, info))
    {
        res= GetLastError();
        CloseHandle(hDsk);
        return res;
    }

    if (pbInput->dwOffset % info.di_bytes_per_sect) {
        debug("ERROR: ITWriteSDCard : startoffset not on sector boundary: sector nr %x ofs %04x\n",
            pbInput->dwOffset / info.di_bytes_per_sect, pbInput->dwOffset % info.di_bytes_per_sect);
        CloseHandle(hDsk);
        return ERROR_INVALID_PARAMETER;
    }
    if ((pbInput->dwOffset+pbInput->dwSize) % info.di_bytes_per_sect) {
        debug("ERROR: ITWriteSDCard : endoffset not on sector boundary: sector nr %x ofs %04x\n",
            (pbInput->dwOffset+pbInput->dwSize)/ info.di_bytes_per_sect, (pbInput->dwOffset+pbInput->dwSize)% info.di_bytes_per_sect);
        CloseHandle(hDsk);
        return ERROR_INVALID_PARAMETER;
    }

    WriteSDCardResult *out= (WriteSDCardResult *)LocalAlloc(LPTR, sizeof(WriteSDCardResult));
    if (!DiskWriteBlocks(hDsk, pbInput->dwOffset / info.di_bytes_per_sect, pbInput->dwSize / info.di_bytes_per_sect, info, pbInput->buffer, &out->dwNumberOfBytesWritten))
    {
        res= GetLastError();
        CloseHandle(hDsk);
        LocalFree(out);
        return res;
    }
    CloseHandle(hDsk);
    *ppbOutput= out;
    *pcbOutput= sizeof(WriteSDCardResult);

    return 0;
}

//--------------------------------------------------------
DWORD DoMagicianFlashRom(DWORD dwOffset, DWORD dwSize, DWORD dwChecksum, BYTE *pBuffer)
{
    struct FlashAccessCommand {
        DWORD dwCommand;
        DWORD pbDestAddr;
        DWORD pbSrcAddr;
        DWORD dwLength;
    } cmd;

    cmd.dwCommand= 0;
    cmd.pbDestAddr= dwOffset;
    cmd.pbSrcAddr= (DWORD)pBuffer;
    cmd.dwLength= dwSize;

    DWORD dwReturned;
    DWORD res= 0;
    debug("flashrom: erase %d buf=%08lx dst=%08lx len=%08lx\n", cmd.dwCommand, cmd.pbSrcAddr, cmd.pbDestAddr, cmd.dwLength);
    if (!KernelIoControl(0x01012a10, &cmd, sizeof(cmd), NULL, 0, &dwReturned)) {
        res= GetLastError();
        if (res==0)
            res= ERROR_GEN_FAILURE;
        debug("flashaccess: erase: error=%08lx: %08lx\n", res, dwReturned);
        return res;
    }

    cmd.dwCommand= 1;
    cmd.pbDestAddr= dwOffset;
    cmd.pbSrcAddr= (DWORD)pBuffer;
    cmd.dwLength= dwSize;

    debug("flashrom: write %d buf=%08lx dst=%08lx len=%08lx\n", cmd.dwCommand, cmd.pbSrcAddr, cmd.pbDestAddr, cmd.dwLength);
    if (!KernelIoControl(0x01012a10, &cmd, sizeof(cmd), NULL, 0, &dwReturned)) {
        res= GetLastError();
        if (res==0)
            res= ERROR_GEN_FAILURE;
        debug("flashaccess: write: error=%08lx: %08lx\n", res, dwReturned);
        return res;
    }


    return res;
}

#define IOCTL_HAL_OAK_START                 5000
#define IOCTL_HAL_XIP_BASE                  IOCTL_HAL_OAK_START + 500

/*++

	IOCTL_HAL_WRITE_XIP

	Writes an XIP region to flash
	
	Input Buffer is a WRITE_XIP_HEADER followed immediately by dwNumRecords
    of XIP_RECORD_INFO structures

	Output buffer is unused.

--*/


#define IOCTL_HAL_WRITE_XIP_3   CTL_CODE(FILE_DEVICE_HAL, IOCTL_HAL_XIP_BASE + 1, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define IOCTL_HAL_WRITE_XIP_4   CTL_CODE(FILE_DEVICE_HAL, 0x2e, METHOD_BUFFERED, FILE_ANY_ACCESS)

        
typedef struct _WRITE_XIP_HEADER
{
	LPVOID  pvPhysAddr;     // physical starting address of XIP
    DWORD	dwLength;       // length of XIP
    DWORD   dwNumRecords;   // number of records in XIP

} WRITE_XIP_HEADER, *PWRITE_XIP_HEADER;

typedef struct _XIP_RECORD_INFO {

    LPVOID  pvPhysStart;    // physical starting address of record, dest addr
    DWORD   dwPhysLen;      // physical len, in bytes of record
    DWORD   dwCheckSum;     // sum of all the bytes in the record data
    VOID UNALIGNED * pvRAMStart;     // virtual address of record in RAM, source addr

}   XIP_RECORD_INFO, *PXIP_RECORD_INFO;
DWORD DoXipFlashRom(DWORD dwOffset, DWORD dwSize, DWORD dwChecksum, BYTE *pBuffer)
{
    struct {
        WRITE_XIP_HEADER hdr;
        XIP_RECORD_INFO info;
    } xipcmd;

    xipcmd.hdr.pvPhysAddr= (void*)dwOffset;
    xipcmd.hdr.dwLength= dwSize;
    xipcmd.hdr.dwNumRecords= 1;
    xipcmd.info.pvPhysStart= (void*)dwOffset;
    xipcmd.info.dwPhysLen= dwSize;
    xipcmd.info.dwCheckSum= dwChecksum;
    xipcmd.info.pvRAMStart= (VOID*)pBuffer;

    DWORD ioctlcode;
    if (isWince3x())
        ioctlcode=IOCTL_HAL_WRITE_XIP_3;
    else
        ioctlcode=IOCTL_HAL_WRITE_XIP_4;

    DWORD dwResult;
    DWORD dwReturned;
    DWORD res= 0;
    debug("flashrom: csum=%08lx ioctl=%08lx buf=%08lx dst=%08lx len=%08lx\n",
            xipcmd.info.dwCheckSum, ioctlcode, xipcmd.info.pvRAMStart, xipcmd.info.pvPhysStart, xipcmd.info.dwPhysLen);
    if (!KernelIoControl(ioctlcode, &xipcmd, sizeof(xipcmd), &dwResult, sizeof(DWORD), &dwReturned)) {
        res= GetLastError();
        if (res==0)
            res= ERROR_GEN_FAILURE;
    }
    debug("flashxip: error=%08lx: %08lx %08lx\n", res, dwResult, dwReturned);

    return res;
}

ITSUTILS_API HRESULT STDAPICALLTYPE ITFlashROM(
		DWORD cbInput, FlashROMParams *pbInput,
		DWORD *pcbOutput, FlashROMResult **ppbOutput,
		IRAPIStream *pStream)
{
    *pcbOutput= sizeof(FlashROMResult);
    FlashROMResult *pOut= *ppbOutput= (FlashROMResult*)LocalAlloc(LPTR, *pcbOutput);

    BOOL bMode = SetKMode(TRUE);
    DWORD dwPerm = SetProcPermissions(0xFFFFFFFF);

    DWORD res= 0;
    switch (pbInput->dwType)
    {
        case FLASH_TYPE_XIPFLASH:
            res= DoXipFlashRom(pbInput->dwOffset, pbInput->dwSize, pbInput->dwCheckSum, pbInput->buffer);
            break;
        case FLASH_TYPE_MAGICIAN:
            res= DoMagicianFlashRom(pbInput->dwOffset, pbInput->dwSize, pbInput->dwCheckSum, pbInput->buffer);
            break;
        default:
            debug("ERROR: unknown flash type %d\n", pbInput->dwType);
    }
    pOut->dwResult= res;

    debug("flashrom: error=%08lx: %08lx\n", res, pOut->dwResult);
    SetProcPermissions(dwPerm);
    SetKMode(bMode);

    return res;
}

//------------------------------------------------------------------------------------

#define GSM_READ_BLOCK_SIZE 100
bool ReadGsmMemory(DWORD dwStart, BYTE *buf, DWORD dwLength, DWORD *pdwCopied)
{
    DWORD nCopied= 0;
    DWORD addr= dwStart;
    GsmDevice gsm;
    gsm.open();
    gsm.DisableRIL();

    while (nCopied < dwLength)
    {
        DWORD nChunkSize= min(GSM_READ_BLOCK_SIZE -(addr%GSM_READ_BLOCK_SIZE), dwLength-nCopied);
        if (!gsm.ReadBlock(&buf[nCopied], addr, nChunkSize))
            return false;
        nCopied += nChunkSize;
        addr += nChunkSize;
    }
    *pdwCopied= nCopied;

    return true;
}

#define GSM_WRITE_BLOCK_SIZE 4
bool WriteGsmMemory(DWORD dwStart, BYTE *buf, DWORD dwLength, DWORD *pdwCopied)
{
    DWORD nCopied= 0;
    DWORD addr= dwStart;
    GsmDevice gsm;
    gsm.open();

    while (nCopied < dwLength)
    {
        DWORD nChunkSize= min(GSM_WRITE_BLOCK_SIZE -(addr%GSM_WRITE_BLOCK_SIZE), dwLength-nCopied);

        if (!gsm.WriteBlock(addr, &buf[nCopied], nChunkSize))
            return false;
        nCopied += nChunkSize;
        addr += nChunkSize;
    }
    *pdwCopied= nCopied;
    return true;
}

ITSUTILS_API HRESULT STDAPICALLTYPE ITReadGsmMemory(
        DWORD cbInput, ReadGsmMemoryParams *pbInput,
        DWORD *pcbOutput, ReadGsmMemoryResult **ppbOutput,
        IRAPIStream *pStream)
{
    ReadGsmMemoryResult *pOut=NULL;

    *pcbOutput= PTR_DIFF(pOut, pOut->buffer+pbInput->nSize);

    pOut= *ppbOutput= (ReadGsmMemoryResult*)LocalAlloc(LPTR, *pcbOutput);

    DWORD res= 0;
    if (!ReadGsmMemory(pbInput->dwOffset, &pOut->buffer[0], pbInput->nSize, &pOut->dwNumberOfBytesRead))
        res= GetLastError();

    return res;
}

ITSUTILS_API HRESULT STDAPICALLTYPE ITWriteGsmMemory(
        DWORD cbInput, WriteGsmMemoryParams *pbInput,
        DWORD *pcbOutput, WriteGsmMemoryResult **ppbOutput,
        IRAPIStream *pStream)
{
    *pcbOutput= sizeof(WriteGsmMemoryResult);
    WriteGsmMemoryResult *pOut= *ppbOutput= (WriteGsmMemoryResult*)LocalAlloc(LPTR, *pcbOutput);

    DWORD res= 0;
    if (!WriteGsmMemory(pbInput->dwOffset, &pbInput->buffer[0], pbInput->nSize, &pOut->dwNumberOfBytesWritten))
        res= GetLastError();

    return res;
}

void GsmReset()
{
    GsmDevice gsm;
    gsm.open();
    gsm.DisableRIL();
    gsm.Reset();
}

void GsmStopRil()
{
    GsmDevice gsm;
    gsm.open();
    gsm.DisableRIL();
    gsm.close();
}

void GsmStartRil()
{
    GsmDevice gsm;
    gsm.open();
    gsm.EnableRIL();
    gsm.close();
}

ITSUTILS_API HRESULT STDAPICALLTYPE ITGsmControl(
        DWORD cbInput, GsmControlParams *pbInput,
        DWORD *pcbOutput, GsmControlResult **ppbOutput,
        IRAPIStream *pStream)
{
    *pcbOutput= 0;
    *ppbOutput= NULL;

    switch(pbInput->nCmd)
    {
        case GSM_RESET: GsmReset(); break;
        case GSM_STOPRIL: GsmStopRil(); break;
        case GSM_STARTRIL: GsmStartRil(); break;
    }

    return 0;
}


ITSUTILS_API HRESULT STDAPICALLTYPE ITReboot(
        DWORD cbInput, BYTE *pbInput,
        DWORD *pcbOutput, BYTE **ppbOutput,
        IRAPIStream *pStream)
{
    *pcbOutput= 0;
    *ppbOutput= NULL;

    KernelIoControl(IOCTL_HAL_REBOOT, NULL, 0, NULL, 0, NULL);

    return 0;
}


ITSUTILS_API HRESULT STDAPICALLTYPE ITSetSystemTime(
        DWORD cbInput, SetSystemTimeParams *pbInput,
        DWORD *pcbOutput, BYTE **ppbOutput,
        IRAPIStream *pStream)
{
    *pcbOutput= 0;
    *ppbOutput= NULL;

    DWORD res= 0;
    BOOL bMode = SetKMode(TRUE);
    DWORD dwPerm = SetProcPermissions(0xFFFFFFFF);

    if (!SetTimeZoneInformation(&pbInput->tzinfo))
        res= GetLastError();

    if (!SetSystemTime(&pbInput->time))
        res= GetLastError();

    SetProcPermissions(dwPerm);
    SetKMode(bMode);

    return res;
}
ITSUTILS_API HRESULT STDAPICALLTYPE ITFindWindow(
        DWORD cbInput, FindWindowParams *pbInput,
        DWORD *pcbOutput, FindWindowResult **ppbOutput,
        IRAPIStream *pStream)
{
    *pcbOutput= sizeof(FindWindowResult);
    FindWindowResult *pOut= *ppbOutput= (FindWindowResult*)LocalAlloc(LPTR, *pcbOutput);

    pOut->hWnd= FindWindow(pbInput->szClassName[0]?pbInput->szClassName:NULL, pbInput->szWindowName[0]?pbInput->szWindowName:NULL);
    if (pOut->hWnd==NULL)
        return GetLastError();

    return 0;
}

ITSUTILS_API HRESULT STDAPICALLTYPE ITSendMessage(
        DWORD cbInput, SendMessageParams *pbInput,
        DWORD *pcbOutput, SendMessageResult **ppbOutput,
        IRAPIStream *pStream)
{
    debug("ITSendMessage(%08lx %08lx %08lx %08lx %08lx %08lx, %d)\n",
            pbInput->dwFlags, pbInput->dwResultAlloc, pbInput->hWnd,
            pbInput->nMsg, pbInput->wParam, pbInput->lParam, 
            cbInput-PTR_DIFF(pbInput, pbInput->buf));

    SendMessageResult *pOut = NULL;
    *pcbOutput= PTR_DIFF(pOut, pOut->buf+pbInput->dwResultAlloc);
    pOut= *ppbOutput= (SendMessageResult*)LocalAlloc(LPTR, *pcbOutput);
    if (pOut==NULL)
        return ERROR_NOT_ENOUGH_MEMORY;

    WPARAM wParam = pbInput->wParam;
    if (pbInput->dwFlags&SMGS_WPARAM_IS_INPUTOFS)
        wParam += (DWORD)pbInput;
    else if (pbInput->dwFlags&SMGS_WPARAM_IS_OUTPUTOFS)
        wParam += (DWORD)pOut;

    LPARAM lParam = pbInput->lParam;
    if (pbInput->dwFlags&SMGS_LPARAM_IS_INPUTOFS)
        lParam += (DWORD)pbInput;
    else if (pbInput->dwFlags&SMGS_LPARAM_IS_OUTPUTOFS)
        lParam += (DWORD)pOut;

    BOOL bMode = SetKMode(TRUE);
    DWORD dwPerm = SetProcPermissions(0xFFFFFFFF);

    DWORD res= 0;
    if (pbInput->dwFlags&SMGS_POSTMESSAGE) {
        debug("PostMessage(%08lx, %08lx, %08lx, %08lx) [%x]\n",
                pbInput->hWnd, pbInput->nMsg, wParam, lParam, pbInput->dwFlags&15);
        if (!PostMessage(pbInput->hWnd, pbInput->nMsg, wParam, lParam))
            res= GetLastError();
    }
    else {
        debug("SendMessage(%08lx, %08lx, %08lx, %08lx) [%x]\n",
                pbInput->hWnd, pbInput->nMsg, wParam, lParam, pbInput->dwFlags&15);
        pOut->lResult= SendMessage(pbInput->hWnd, pbInput->nMsg, wParam, lParam);
    }

    SetProcPermissions(dwPerm);
    SetKMode(bMode);

    return res;
}

ITSUTILS_API HRESULT STDAPICALLTYPE ITCallAddress(
        DWORD cbInput, CallAddressParams *pbInput,
        DWORD *pcbOutput, BYTE **ppbOutput,
        IRAPIStream *pStream)
{
    *pcbOutput= 0;
    *ppbOutput= NULL;

    typedef void (*calltarget_fn)();

    calltarget_fn  calltarget= (calltarget_fn)pbInput->dwAddress;

    BOOL bMode = SetKMode(TRUE);
    DWORD dwPerm = SetProcPermissions(0xFFFFFFFF);

    // todo: if (pbInput->hProcess==INVALID_HANDLE_VALUE) -> call physical address
    // if (pbInput->hProcess==NULL) -> call virtual address
    // else call to address in process space
    calltarget();

    SetProcPermissions(dwPerm);
    SetKMode(bMode);

    return 0;
}

