/* (C) 2003 XDA Developers  itsme@xs4all.nl
 *
 * $Header: /var/db/cvs/xda-devtools/itsutils/leds/tststrata.cpp,v 1.1 2005/06/12 23:09:03 itsme Exp $
 */
#include  <windows.h>
#include  <pm.h>
#include "pkfuncs.h"
#include "debug.h"
#include "stringutils.h"
#include "vectorutils.h"

// experimental program, to read the intel strataflash chip information
//
const char* storagepath= "\\Storage Card\\dat";
//  partition size seems to be 0x400000
//  0000-000f  appear every 0x0800 bytes
//  0010-0400  appear every 0x2000 bytes
template<class TVector>
void WriteVector(const TVector& data, const std::string& name)
{
    if (data.empty())
        return;
    FILE *f= fopen(name.c_str(), "a+b");
    if (f) {
        fwrite(vectorptr(data), sizeof(TVector::value_type), data.size(), f);
        fclose(f);
    }
}


extern "C"
BOOL KernelIoControl(DWORD dwIoControlCode, LPVOID lpInBuf, DWORD nInBufSize, LPVOID lpOutBuf, DWORD nOutBufSize, LPDWORD lpBytesReturned);

inline void flashcmd(DWORD addr, BYTE cmd) {
    *(DWORD*)addr= cmd|(cmd<<16);
}
inline WORD readword(DWORD addr) {
    return *(WORD*)addr;
}
inline BYTE readbyte(DWORD addr) {
    return *(BYTE*)addr;
}
inline DWORD readdword(DWORD addr) {
    return *(DWORD*)addr;
}
inline void disable_irq()
{
    if (!KernelIoControl(0x01012a24, 0, 0, 0, 0, 0))
        error("disable_irq");
}
inline void enable_irq()
{
    if (!KernelIoControl(0x01012a20, 0, 0, 0, 0, 0))
        error("enable_irq");
}
inline read_status(DWORD dwPartition)
{
    // addr can be any address in the partition
    flashcmd(dwPartition, 0x70);
    return readbyte(dwPartition);
}
inline clear_status(DWORD dwAddr)
{
    // addr can be any address in the flash
    flashcmd(dwAddr, 0x50);
}
inline set_read_array_mode(DWORD dwPartition)
{
    // addr can be any address in the partition
    flashcmd(dwPartition, 0xff);
}
inline set_read_cfi_mode(DWORD dwPartition)
{
    // addr can be any address in the partition
    flashcmd(dwPartition, 0x98);
}
inline set_read_di_mode(DWORD dwPartition)
{
    // addr can be any address in the partition
    flashcmd(dwPartition, 0x90);
}
//////////////////////////////////////////////////////////////////////////
//
// write buffered:
// flashcmd(dwAddr, 0xe8)
// wait for SR.7  'write buffer available'
// flashcmd(dwAddr, data.size()) 
// for (i=0; i<data.size() ; i++)
//    flashcmd(dwAddr+i, data[i]);
// flashcmd(dwAddr, 0xd0)
// wait for SR.7  'buffered programming finished'

//////////////////////////////////////////////////////////////////////////

inline read_cfi_data(DWORD dwPartition, DWORD dwReg)
{
    // read from cfi address + offset
    return readword(dwPartition+dwReg*2);
}
void flash_read_cfi(DWORD dwPartition, DWORD dwLength)
{

    debug("start readcfi(%08lx)\n", dwPartition);

    WordVector surround;  surround.resize(24);
    WordVector regs;  regs.resize(dwLength/2);

    disable_irq();

    clear_status(dwPartition);
    DWORD status1= read_status(dwPartition);

    set_read_cfi_mode(dwPartition);

    int i;
    if (dwPartition-0x800>=0xa0000000)
        for (i= 0; i < 8; i++)
            surround[i]= read_cfi_data(dwPartition, i-0x800);

    if (dwPartition-8>=0xa0000000)
        for (i= 0; i < 8; i++)
            surround[i+8]= read_cfi_data(dwPartition, i-8);

    for (DWORD dwReg= 0; dwReg < regs.size(); dwReg++)
        regs[dwReg]= read_cfi_data(dwPartition, dwReg);

    if (dwPartition+regs.size()<0xa4000000)
        for (i= 0 ; i < 8; i++) 
            surround[i+16]= read_cfi_data(dwPartition, i+regs.size());

    BYTE status2= read_status(dwPartition);

    set_read_array_mode(dwPartition);

    enable_irq();

    WriteVector(regs, stringformat("%s\\cfi-%08lx.dat", storagepath, dwPartition));
    debug("done readcfi(%08lx) %02x %02x : %hs\n", dwPartition, status1, status2, 
            hexdump((BYTE*)vectorptr(surround), surround.size(), 2).c_str());
}
//////////////////////////////////////////////////////////////////////////

inline read_di_data(DWORD dwPartition, DWORD dwReg)
{
    // read from di address + offset
    return readword(dwPartition+dwReg*2);
}
void flash_read_di(DWORD dwPartition, DWORD dwLength)
{

    debug("start readdi(%08lx)\n", dwPartition);

    WordVector surround;  surround.resize(24);
    WordVector regs;  regs.resize(dwLength/2);

    disable_irq();

    clear_status(dwPartition);
    DWORD status1= read_status(dwPartition);

    set_read_di_mode(dwPartition);

    int i;
    if (dwPartition-0x800>=0xa0000000)
        for (i= 0; i < 8; i++)
            surround[i]= read_di_data(dwPartition, i-0x800);

    if (dwPartition-8>=0xa0000000)
        for (i= 0; i < 8; i++)
            surround[i+8]= read_di_data(dwPartition, i-8);

    for (DWORD dwReg= 0; dwReg < regs.size(); dwReg++)
        regs[dwReg]= read_di_data(dwPartition, dwReg);

    if (dwPartition+regs.size()<0xa4000000)
        for (i= 0 ; i < 8; i++) 
            surround[i+16]= read_di_data(dwPartition, i+regs.size());

    BYTE status2= read_status(dwPartition);

    set_read_array_mode(dwPartition);

    enable_irq();

    WriteVector(regs, stringformat("%s\\di-%08lx.dat", storagepath, dwPartition));
    debug("done readdi(%08lx) %02x %02x : %hs\n", dwPartition, status1, status2, 
            hexdump((BYTE*)vectorptr(surround), surround.size(), 2).c_str());
}



int WINAPI WinMain( HINSTANCE hInstance,
                   HINSTANCE hPrevInstance,
                   LPTSTR    lpCmdLine,
                   int       nCmdShow)
{
    DebugSetLogfile("tststrata.log");
    DebugOutputDebugString();

    debug("starting tststrata.log: %08lx\n", GetTickCount());

    BOOL bMode = SetKMode(TRUE);
    DWORD dwPerm = SetProcPermissions(0xFFFFFFFF);
/*
    for (DWORD dwAddr= 0xa0000000 ; dwAddr<0xa4000000 ; dwAddr += 0x8000)
        flash_read_cfi(dwAddr, 0x8000);
*/
    for (DWORD dwAddr= 0xa0000000 ; dwAddr<0xa4000000 ; dwAddr += 0x8000)
        flash_read_di(dwAddr, 0x8000);

    SetProcPermissions(dwPerm);
    SetKMode(bMode);

    debug("ending tststrata.log: %08lx\n", GetTickCount());

    return 0;
}

