/* (C) 2003 XDA Developers  itsme@xs4all.nl
 *
 * $Header: /var/db/cvs/xda-devtools/itsutils/src/checkdll.cpp,v 1.9 2005/06/12 22:53:03 itsme Exp $
 */
#include <stdlib.h>
#include <windows.h>
#include <rapi.h>

#include <string>
#include "vectorutils.h"
#include "stringutils.h"

#include "debug.h"
#include "dllversion.h"
char *GetAppPath()
{
    static char appname[1024];
    if (!GetModuleFileName(NULL, appname, 1024))
        return NULL;

    return appname;
}

bool CeFileExists(const std::string& name)
{
    DWORD dwAttr = CeGetFileAttributes( ToWString(name).c_str() );
    if (0xFFFFFFFF == dwAttr)
        return false;

    return true;
}
bool CeCopyFileToDevice(const std::string& srcfile, const std::string& dstfile) 
{
    WIN32_FIND_DATA wfd;
    HANDLE hFind = FindFirstFile( srcfile.c_str(), &wfd);
    if (INVALID_HANDLE_VALUE == hFind)
    {
        error("Source/host file does not exist '%hs'\n", srcfile.c_str());
        return false;
    }
    FindClose( hFind);

    if (wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
    {
        error("Source/host file specifies a directory '%hs'\n", srcfile.c_str());
        return false;
    }

    std::string dstname= dstfile;

    DWORD dwAttr = CeGetFileAttributes( ToWString(dstname).c_str());
    if (0xFFFFFFFF  != dwAttr)
    {
        if (dwAttr & FILE_ATTRIBUTE_DIRECTORY)
        {
            dstname += "\\";
            dstname += wfd.cFileName;
        }
        else
        {
            ceerror("File already exists. '%hs'\n", dstname.c_str());
            return false;
        }
    }

    HANDLE hSrc = CreateFile( srcfile.c_str(), GENERIC_READ, FILE_SHARE_READ,
                NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (INVALID_HANDLE_VALUE == hSrc)
    {
        error("Unable to open source/host file '%hs'", srcfile.c_str());
        return false;
    }

    HANDLE hDest = CeCreateFile( ToWString(dstname).c_str(), GENERIC_WRITE, FILE_SHARE_READ,
                NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
    if (INVALID_HANDLE_VALUE == hDest )
    {
        ceerror("Unable to open WinCE file '%hs'", dstname.c_str());
        return false;
    }

    debug("Copying %hs to WCE:%ls\n", srcfile.c_str(), ToWString(dstname).c_str());

    ByteVector buffer; buffer.resize(32768);

    DWORD dwNumRead;
    do
    {
        if (ReadFile( hSrc, vectorptr(buffer), (DWORD)buffer.size(), &dwNumRead, NULL))
        {
            DWORD dwNumWritten;
            if (!CeWriteFile( hDest, vectorptr(buffer), dwNumRead, &dwNumWritten, NULL))
            {
                ceerror("Error !!! Writing WinCE file");
                CeCloseHandle( hDest);
                CloseHandle (hSrc);
                return false;
            }
        }
        else
        {
            error("Error !!! Reading source file");
            CeCloseHandle( hDest);
            CloseHandle (hSrc);
            return false;
        }
    } while (dwNumRead);
    CeCloseHandle( hDest);
    CloseHandle (hSrc);
    return true;
}
void CheckITSDll()
{
    if (FAILED(CeRapiInit()))
    {
        error("CeRapiInit failed\n");
        exit(1);
    }

    DWORD outsize=0;
    HRESULT res= CeRapiInvoke(L"\\Windows\\ItsUtils.dll", L"ITGetVersion",
            0, NULL, &outsize, NULL, NULL, 0);
    // E_FAIL  -> itsutils.dll is missing, or execute denied by policy
    if (res==E_FAIL && CeFileExists("\\Windows\\ItsUtils.dll")) {
        debug("ERROR loading itsutils.dll - probably denied by policy restriction\n");
        exit(1);
    }
	if (res==0x80072745 || res==0x80072746) {
		error(res, "CeRapiInvoke");
		debug("\nplease reboot your device, or restart activesync\n");
		exit(1);
	}
    if (res!=ITSDLL_VERSION)
    {
        std::string approot= GetAppPath();
	    approot.resize(approot.find_last_of('\\'));
		if (!CeCopyFileToDevice(approot+"\\itsutils.dll", "\\windows"))
			exit(1);

		res= CeRapiInvoke(L"\\Windows\\ItsUtils.dll", L"ITGetVersion",
            0, NULL, &outsize, NULL, NULL, 0);
		if (res==E_FAIL && CeFileExists("\\Windows\\ItsUtils.dll")) {
			debug("ERROR loading itsutils.dll - probably denied by policy restriction\n");
			exit(1);
		}
    }
	if (res!=ITSDLL_VERSION) {
		debug("Could not update itsutils.dll to the current version, maybe it is inuse?\n");
		debug("try restarting your device, or restart activesync\n");
		exit(1);
	}
}
