/* (C) 2003 XDA Developers  itsme@xs4all.nl
 *
 * $Header: /var/db/cvs/xda-devtools/itsutils/src/pmemmap.cpp,v 1.11 2005/06/27 11:21:34 itsme Exp $
 */
#include <windows.h>
#include "ItsUtils.h"
#include "dllversion.h"
#include "debug.h"
#include "args.h"
#include <stdio.h>
#include <string.h>
#include <rapi.h>

#include <string>
#include "stringutils.h"

// todo: add option to save all sections to files.
// todo: add option to produce list of  v -> p, p, p ...
// todo: create tool to search sections for binary string.
bool ITReadProcessMemory(HANDLE hProc, DWORD dwOffset, BYTE *buffer, DWORD dwBytesWanted, DWORD *pdwNumberOfBytesRead);

DWORD dwPhys_min= 0x10000000;
DWORD dwPhys_max= 0x18000000;
std::string g_savefilename;
FILE *g_f= NULL;
bool is_in_save_range(DWORD dwPhys)
{
    return (dwPhys_min<=dwPhys && dwPhys<dwPhys_max);
}
bool save_to_section_file(DWORD dwStart, DWORD dwLength)
{
    if (g_savefilename.empty())
        return false;

    DWORD nRead;
    ByteVector data; data.resize(dwLength);
    if (!ITReadProcessMemory(NULL, dwStart, vectorptr(data), data.size(), &nRead))
        return false;

    if (g_f==NULL) {
        g_f= fopen(g_savefilename.c_str(), "wb");
        if (g_f==NULL) {
            perror(stringformat("open %s", g_savefilename.c_str()).c_str());
            g_savefilename.clear();
            return false;
        }
    }
    if (0!=fseek(g_f, dwStart & 0x1ffffff, SEEK_SET))
        perror(stringformat("seek %08lx %s", dwStart & 0x1ffffff, g_savefilename.c_str()).c_str());
    size_t n= fwrite(vectorptr(data), 1, data.size(), g_f);
    if (n!=data.size())
        perror(stringformat("writing %s", g_savefilename.c_str()).c_str());

    return true;
}

bool ITReadProcessMemory(HANDLE hProc, DWORD dwOffset, BYTE *buffer, DWORD dwBytesWanted, DWORD *pdwNumberOfBytesRead)
{
    ReadProcessMemoryParams inbuf;
    DWORD outsize=0;
    ReadProcessMemoryResult *outbuf=NULL;

    inbuf.hProcess= hProc;
    inbuf.dwOffset= dwOffset;
    inbuf.nSize= dwBytesWanted;
    outbuf= NULL; outsize= 0;
    HRESULT res= CeRapiInvoke(L"\\Windows\\ItsUtils.dll", L"ITReadProcessMemory",
            sizeof(ReadProcessMemoryParams), (BYTE*)&inbuf,
            &outsize, (BYTE**)&outbuf, NULL, 0);
    if (res || outbuf==NULL)
    {
        error(res, "ITReadProcessMemory");
        return false;
    }
    memcpy(buffer, &outbuf->buffer, outbuf->dwNumberOfBytesRead);
    *pdwNumberOfBytesRead= outbuf->dwNumberOfBytesRead;
    LocalFree(outbuf);

    return true;
}
DWORD ReadDword(DWORD dwOffset)
{
    DWORD nRead;
    DWORD dwValue;
    if (!ITReadProcessMemory(NULL, dwOffset, (BYTE*)&dwValue, sizeof(DWORD), &nRead))
        return 0;
    return dwValue;
}

DWORD FindVirtForPhys(const DWORD *ptbl, DWORD dwPhys)
{
    DWORD dwAddr= 0;
    DWORD pidx= 0;
    do {
        DWORD pe= ptbl[pidx];

        if ((pe&3)==2) {
            if ((pe&0xfff00000)==(dwPhys&0xfff00000)) {
                return dwAddr | (dwPhys&0xfffff);
            }
        }

        dwAddr+=0x100000;
        pidx++;
    } while (dwAddr);
    return 0;
}
bool SummarizeTLB2(const DWORD *ptbl1, DWORD dwVirtbase, DWORD dwTlbPhysAddr, DWORD pagesize, int nentries)
{
    DWORD ptbl2ofs= FindVirtForPhys(ptbl1, dwTlbPhysAddr);
    if (ptbl2ofs==NULL) {
        debug("ERROR: could not find virtual address for phys %08lx\n", dwTlbPhysAddr);
        return false;
    }
    DwordVector ptbl2; ptbl2.resize(nentries);
    DWORD nRead;
    if (!ITReadProcessMemory(NULL, ptbl2ofs, (BYTE*)vectorptr(ptbl2), ptbl2.size()*sizeof(DWORD), &nRead))
        return false;

    bool bStarted= false;
    DWORD dwStartVirt= 0;
    DWORD dwStartPhys= 0;
    DWORD dwPrevPhys= 0;

    DWORD dwAddr= dwVirtbase;
    DWORD pidx= 0;
    for (int i=0 ; i<nentries ; i++)
    {
        DWORD pe= ptbl2[pidx];

        DWORD dwCurPhys= 0;
        bool bHasPhysAddr= false;
        bool bEndBlock= false;

        switch (pe&3) {
            case 0:  // ignore entry
                bEndBlock= true;
                break;
            case 1:  // large page descriptor
                dwCurPhys= pe&0xffff0000;
                bHasPhysAddr= true;
                break;
            case 2:  // small page descriptor
                dwCurPhys= pe&0xfffff000;
                bHasPhysAddr= true;
                break;
            case 3:  // tiny page descriptor
                dwCurPhys= pe&0xfffffc00;
                bHasPhysAddr= true;
                break;
        }
        if (dwCurPhys-dwPrevPhys!=pagesize)
            bEndBlock= true;

        if (bStarted && bEndBlock) {
            debug("v%08lx-%08lx -> p%08lx-%08lx\n", dwStartVirt, dwAddr, dwStartPhys, dwPrevPhys+pagesize);
            bStarted= false;
        }
        if (bHasPhysAddr) {
            if (!bStarted) {
                bStarted= true;
                dwStartPhys= dwCurPhys;
                dwStartVirt= dwAddr;
            }
            dwPrevPhys= dwCurPhys;
        }

        dwAddr+=pagesize;
        pidx++;
    }
    if (bStarted) {
        debug("v%08lx-%08lx -> p%08lx-%08lx\n", dwStartVirt, dwAddr, dwStartPhys, dwPrevPhys+pagesize);
        bStarted= false;
    }

    return true;
}
bool SummarizePageTable(bool bVerbose)
{
    DWORD ptbl1ofs= 0xfffd0000;
    DwordVector ptbl1; ptbl1.resize(4096);
    DWORD nRead;
    if (!ITReadProcessMemory(NULL, ptbl1ofs, (BYTE*)vectorptr(ptbl1), ptbl1.size()*sizeof(DWORD), &nRead))
        return false;

    DWORD pagesize= 0x100000;

    bool bStarted= false;
    DWORD dwStartVirt= 0;
    DWORD dwStartPhys= 0;
    DWORD dwPrevPhys= 0;

    DWORD dwAddr= bVerbose ? 0 : 0x80000000;
    DWORD pidx= dwAddr/pagesize;
    do {
        DWORD pe= ptbl1[pidx];

        DWORD dwCurPhys= 0;
        DWORD dwTlbPhys= 0;
        DWORD dwTlbPagesize= 0;
        bool bHasPhysAddr= false;
        bool bEndBlock= false;
        bool bHasTlb2= false;

        switch (pe&3) {
            case 0:  // ignore entry
                bEndBlock= true;
                break;
            case 1:  // coarse tlb entry
                dwTlbPhys= pe&0xfffffc00;
                dwTlbPagesize= 0x1000;
                bHasTlb2= true;
                bEndBlock= true;
                break;
            case 2:  // section descriptor
                dwCurPhys= pe&0xfff00000;
                bHasPhysAddr= true;
                break;
            case 3:  // fine tlb entry
                dwTlbPhys= pe&0xfffff000;
                dwTlbPagesize= 0x400;
                bHasTlb2= true;
                bEndBlock= true;
                break;
        }
        if (dwCurPhys-dwPrevPhys!=pagesize)
            bEndBlock= true;

        if (bStarted && bEndBlock) {
            debug("v%08lx-%08lx -> p%08lx-%08lx\n", dwStartVirt, dwAddr, dwStartPhys, dwPrevPhys+pagesize);
            bStarted= false;
        }
        if (bHasTlb2) {
            SummarizeTLB2(vectorptr(ptbl1), dwAddr, dwTlbPhys, dwTlbPagesize, pagesize/dwTlbPagesize);
        }
        if (bHasPhysAddr) {
            if (!bStarted) {
                bStarted= true;
                dwStartPhys= dwCurPhys;
                dwStartVirt= dwAddr;
            }
            dwPrevPhys= dwCurPhys;
        }

        dwAddr+=pagesize;
        pidx++;
    } while (bVerbose ? dwAddr : dwAddr<0xa0000000);
    if (bStarted) {
        debug("v%08lx-%08lx -> p%08lx-%08lx\n", dwStartVirt, dwAddr, dwStartPhys, dwPrevPhys+pagesize);
        bStarted= false;
    }
    return true;
}

// verbose
bool SummarizeSection(DWORD dwSectionAddr)
{
    DWORD sectionTableOfs= 0xFFFFC8A0;
    DwordVector SectionTable; SectionTable.resize(64);
    DWORD nRead;
    if (!ITReadProcessMemory(NULL, sectionTableOfs, (BYTE*)vectorptr(SectionTable), SectionTable.size()*sizeof(DWORD), &nRead))
        return false;

    DWORD pagesize= 0x1000;

#define VA_BLOCK   16
#define VA_SECTION 25
#define VA_PAGE    12
#define BLOCK_MASK 0x1ff
#define PAGE_MASK  0x00f
#define NULL_BLOCK      0
#define RESERVED_BLOCK  1

    DWORD dwSectionOfs= SectionTable[dwSectionAddr>>VA_SECTION];
    DwordVector section; section.resize(512);
    if (!ITReadProcessMemory(NULL, dwSectionOfs, (BYTE*)vectorptr(section), section.size()*sizeof(DWORD), &nRead))
        return false;

    DWORD dwStartPhys=0;
    DWORD dwStartVirt=0;
    DWORD dwPrevPhys=0;
    bool bStarted= false;
    DWORD dwVAddr= 0;
    for (DWORD ixBlock= 0 ; ixBlock<section.size() ; ixBlock++) {
        DWORD dwMemblockOfs = section[ixBlock];
        dwVAddr= dwSectionAddr+(ixBlock<<VA_BLOCK);

        if (dwMemblockOfs==NULL_BLOCK || dwMemblockOfs==RESERVED_BLOCK) {
            if (bStarted) {
                debug("v%08lx-%08lx -> p%08lx-%08lx\n", dwStartVirt, dwVAddr, dwStartPhys, dwPrevPhys+pagesize);
                bStarted= false;
            }
        }
        else {
            DwordVector memblock; memblock.resize(4);
            if (!ITReadProcessMemory(NULL, dwMemblockOfs, (BYTE*)vectorptr(memblock), memblock.size()*sizeof(DWORD), &nRead))
                return false;
            DWORD dwPagesOfs= memblock[3];
            DwordVector pages; pages.resize(16);
            if (!ITReadProcessMemory(NULL, dwPagesOfs, (BYTE*)vectorptr(pages), pages.size()*sizeof(DWORD), &nRead))
                return false;

            for (DWORD ixPage= 0 ; ixPage<pages.size() ; ixPage++) {
                DWORD dwPageentry= pages[ixPage];
                if (dwPageentry==0xfffffff0 || dwPageentry==0 )  {
                    if (bStarted) {
                        debug("v%08lx-%08lx -> p%08lx-%08lx\n", dwStartVirt, dwVAddr, dwStartPhys, dwPrevPhys+pagesize);
                        bStarted= false;
                    }
                }
                else {
                    DWORD dwPhys= pages[ixPage]&0xfffff000;
                    if (!g_savefilename.empty() && is_in_save_range(dwPhys))
                        save_to_section_file(dwVAddr, pagesize);
                    if (dwPhys-dwPrevPhys!=pagesize) {
                        if (bStarted) {
                            debug("v%08lx-%08lx -> p%08lx-%08lx\n", dwStartVirt, dwVAddr, dwStartPhys, dwPrevPhys+pagesize);
                            bStarted= false;
                        }
                    }
                    if (!bStarted) {
                        bStarted= true;
                        dwStartPhys= dwPhys;
                        dwStartVirt= dwVAddr;
                    }
                    dwPrevPhys= dwPhys;

                }
                dwVAddr+=pagesize;
            }
        }
    }
    if (bStarted) {
        debug("v%08lx-%08lx -> p%08lx-%08lx\n", dwStartVirt, dwVAddr, dwStartPhys, dwPrevPhys+pagesize);
        bStarted= false;
    }
    return true;
}
// only summarize virtual addresses
bool SummarizeSectionVirtual(DWORD dwSectionAddr)
{
    DWORD sectionTableOfs= 0xFFFFC8A0;
    DwordVector SectionTable; SectionTable.resize(64);
    DWORD nRead;
    if (!ITReadProcessMemory(NULL, sectionTableOfs, (BYTE*)vectorptr(SectionTable), SectionTable.size()*sizeof(DWORD), &nRead))
        return false;

    DWORD pagesize= 0x1000;

#define VA_BLOCK   16
#define VA_SECTION 25
#define VA_PAGE    12
#define BLOCK_MASK 0x1ff
#define PAGE_MASK  0x00f
#define NULL_BLOCK      0
#define RESERVED_BLOCK  1

    DWORD dwSectionOfs= SectionTable[dwSectionAddr>>VA_SECTION];
    DwordVector section; section.resize(512);
    if (!ITReadProcessMemory(NULL, dwSectionOfs, (BYTE*)vectorptr(section), section.size()*sizeof(DWORD), &nRead))
        return false;

    DWORD dwStartVirt=0;
    bool bStarted= false;
    DWORD dwVAddr= 0;
    for (DWORD ixBlock= 0 ; ixBlock<section.size() ; ixBlock++) {
        DWORD dwMemblockOfs = section[ixBlock];
        dwVAddr= dwSectionAddr+(ixBlock<<VA_BLOCK);

        if (dwMemblockOfs==NULL_BLOCK || dwMemblockOfs==RESERVED_BLOCK) {
            if (bStarted) {
                debug("v%08lx-%08lx\n", dwStartVirt, dwVAddr);
                bStarted= false;
            }
        }
        else {
            DwordVector memblock; memblock.resize(4);
            if (!ITReadProcessMemory(NULL, dwMemblockOfs, (BYTE*)vectorptr(memblock), memblock.size()*sizeof(DWORD), &nRead))
                return false;
            DWORD dwPagesOfs= memblock[3];
            DwordVector pages; pages.resize(16);
            if (!ITReadProcessMemory(NULL, dwPagesOfs, (BYTE*)vectorptr(pages), pages.size()*sizeof(DWORD), &nRead))
                return false;

            for (DWORD ixPage= 0 ; ixPage<pages.size() ; ixPage++) {
                DWORD dwPageentry= pages[ixPage];
                if (dwPageentry==0xfffffff0 || dwPageentry==0 )  {
                    if (bStarted) {
                        debug("v%08lx-%08lx\n", dwStartVirt, dwVAddr);
                        bStarted= false;
                    }
                }
                else {
                    DWORD dwPhys= pages[ixPage]&0xfffff000;
                    if (!g_savefilename.empty() && is_in_save_range(dwPhys))
                        save_to_section_file(dwVAddr, pagesize);
                    if (!bStarted) {
                        bStarted= true;
                        dwStartVirt= dwVAddr;
                    }

                }
                dwVAddr+=pagesize;
            }
        }
    }
    if (bStarted) {
        debug("v%08lx-%08lx\n", dwStartVirt, dwVAddr);
        bStarted= false;
    }
    return true;
}

void usage()
{
    printf("Usage: pmemmap [-v] [-s SECTIONADDR] [-w savefile] [-p MIN-MAX]\n");
    printf("    -p specifies the range of phys addrs to save to savefile\n");
}
int main( int argc, char *argv[])
{
    bool bVerbose= false;
    DWORD dwSectionAddr=0xFFFFFFFF;

    std::string minmaxstring;

    DebugStdOut();
    for (int i=1 ; i<argc ; i++)
    {
        if (argv[i][0]=='-') switch(argv[i][1])
        {
            case 'v': bVerbose= true; break;
            case 's': HANDLEULOPTION(dwSectionAddr, DWORD); break;
            case 'w': HANDLESTROPTION(g_savefilename); break;
            case 'p': HANDLESTROPTION(minmaxstring); break;
            default: usage(); 
                     return 1;
        }
        else {
            usage();
            return 1;
        }
    }

    if (minmaxstring.size()) {
        size_t idx_dash= minmaxstring.find('-');
        if (idx_dash==minmaxstring.npos) {
            usage();
            return 1;
        }
        dwPhys_min= strtoul(minmaxstring.c_str(), 0, 0);
        dwPhys_max= strtoul(minmaxstring.c_str()+idx_dash+1, 0, 0);
    }
    CheckITSDll();

    if (dwSectionAddr==0xFFFFFFFF)
        SummarizePageTable(bVerbose);
    else if (bVerbose)
        SummarizeSection(dwSectionAddr);
    else
        SummarizeSectionVirtual(dwSectionAddr);

    if (g_f)
        fclose(g_f);

    CeRapiUninit();

    return 0;
}


