/* (C) 2003 XDA Developers  itsme@xs4all.nl
 *
 * $Header: /var/db/cvs/xda-devtools/itsutils/src/pput.cpp,v 1.15 2005/06/23 11:34:44 itsme Exp $
 */

// todo: * support unicode filenames
//       * support %CE1%  etc path indicators
//       * when operating on 1 file, do write error message
#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <rapi.h>
#include "debug.h"
#include "stringutils.h"
#include "vectorutils.h"

bool g_allowOverwrite= false;
bool g_useStdout= false;
bool g_bRecurse= false;

// 0 : no output
// 1 : output unexpected errors
// 2 : output all errors
// 3 : output copied file names
int verbose= 1;

#define AT_NONEXISTANT      1
#define AT_ISDIRECTORY      2
#define AT_ISFILE           3
int getWin32Attributes(const std::string& name)
{
    WIN32_FIND_DATA wfd;
    HANDLE hFind = FindFirstFile( name.c_str(), &wfd);
    if (INVALID_HANDLE_VALUE == hFind)
        return AT_NONEXISTANT;

    FindClose( hFind);

    if (wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
        return AT_ISDIRECTORY;

    return AT_ISFILE;
}
bool isWin32Directory(const std::string& name)
{
    return getWin32Attributes(name)==AT_ISDIRECTORY;
}
bool isWin32File(const std::string& name)
{
    return getWin32Attributes(name)==AT_ISFILE;
}

int getCeAttributes(const std::string& name)
{
    DWORD dwAttr = CeGetFileAttributes( ToWString(name).c_str() );
    if (0xFFFFFFFF == dwAttr)
        return AT_NONEXISTANT;

    if (dwAttr & FILE_ATTRIBUTE_DIRECTORY)
        return AT_ISDIRECTORY;

    return AT_ISFILE;
}
bool isCeDirectory(const std::string& name)
{
    return getCeAttributes(name)==AT_ISDIRECTORY;
}
bool isCeFile(const std::string& name)
{
    return getCeAttributes(name)==AT_ISFILE;
}


bool hasWildCards(const std::string& name)
{
    return name.find_first_of("*?")!=name.npos;
}
std::string concat_path(const std::string& p1, const std::string& p2)
{
    std::string result(p1);

    if (result.size() && result[result.size()-1]!='/' && result[result.size()-1]!='\\')
        result += '\\';

    result += p2;

    return result;
}
std::string GetBasePath(const std::string& name)
{
    size_t lastslash= name.find_last_of("\\/");
    if (lastslash==name.npos)
        return "";

    return name.substr(0, lastslash);
}
std::string GetBaseName(const std::string& name)
{
    size_t lastslash= name.find_last_of("\\/");
    if (lastslash==name.npos)
        return name;

    return name.substr(lastslash+1);
}

bool Win32FileGlob(const std::string& name, StringList& list)
{
    WIN32_FIND_DATA wfd;
    HANDLE hFind = FindFirstFile( name.c_str(), &wfd);
    if (INVALID_HANDLE_VALUE == hFind) {
        if (GetLastError()==ERROR_NO_MORE_FILES)
            return true;
        return false;
    }

    std::string rootpath= GetBasePath(name);
    do {
        if (!(wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
        {
            list.push_back(concat_path(rootpath, wfd.cFileName));
        }

    } while (FindNextFile(hFind, &wfd));

    FindClose(hFind);

    return true;
}
bool CeFileGlob(const std::string& name, StringList& filelist, StringList& dirlist)
{
    CE_FIND_DATA wfd;
    HANDLE hFind = CeFindFirstFile( ToWString(name).c_str(), &wfd);
    if (INVALID_HANDLE_VALUE == hFind) {
        if (CeGetLastError()==ERROR_NO_MORE_FILES)
            return true;
        return false;
    }

    std::string rootpath= GetBasePath(name);
    do {
        if (!(wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
        {
            filelist.push_back(concat_path(rootpath, ToString(wfd.cFileName)));
        }
        else if (ToString(wfd.cFileName)!="." && ToString(wfd.cFileName)!="..")
        {
            dirlist.push_back(concat_path(rootpath, ToString(wfd.cFileName)));
        }

    } while (CeFindNextFile(hFind, &wfd));

    CeFindClose(hFind);

    return true;
}
bool CopyWin32FileToCeFile(const std::string& src, const std::string& dstfile)
{
    if (!g_allowOverwrite && isCeFile(dstfile))
    {
        if (verbose>1) debug("destination file already exists - specify -f to overwrite\n");
        return false;
    }
    HANDLE hSrc = CreateFile( src.c_str(), GENERIC_READ, FILE_SHARE_READ,
                NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (INVALID_HANDLE_VALUE == hSrc)
    {
        if (verbose>1) error("%s", src.c_str());
        return false;
    }

    HANDLE hDest = CeCreateFile( ToWString(dstfile).c_str(), GENERIC_WRITE, FILE_SHARE_READ,
                NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
    if (INVALID_HANDLE_VALUE == hDest )
    {
        if (verbose>1) ceerror("WCE:%s", dstfile.c_str());
        return false;
    }

    if (verbose>2) debug("Copying %hs to WCE:%hs\n", src.c_str(), dstfile.c_str());

    ByteVector buffer;  buffer.resize(65536);

    bool bRes= true;

    DWORD dwNumRead;
    do
    {
        if (!ReadFile( hSrc, vectorptr(buffer), buffer.size(), &dwNumRead, NULL))
        {
            if (verbose>0) error("reading %s", src.c_str());
            bRes= false;
            break;
        }

        DWORD dwNumWritten;
        if (!CeWriteFile( hDest, vectorptr(buffer), dwNumRead, &dwNumWritten, NULL))
        {
            if (verbose>0) ceerror("Writing WCE:%s", dstfile.c_str());
            bRes= false;
            break;
        }
    } while (dwNumRead);

    CeCloseHandle( hDest);
    CloseHandle (hSrc);

    return bRes;
}
bool CopyStdinToCeFile(const std::string& dstfile)
{
    if (!g_allowOverwrite && isCeFile(dstfile))
    {
        if (verbose>1) debug("destination file already exists - specify -f to overwrite\n");
        return false;
    }
    HANDLE hDest = CeCreateFile( ToWString(dstfile).c_str(), GENERIC_WRITE, FILE_SHARE_READ,
                NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
    if (INVALID_HANDLE_VALUE == hDest )
    {
        if (verbose>1) ceerror("WCE:%s", dstfile.c_str());
        return false;
    }

    if (verbose>2) debug("Copying STDIN to WCE:%hs\n", dstfile.c_str());

    ByteVector buffer;  buffer.resize(65536);

    bool bRes= true;

    DWORD dwNumRead;
    while (!feof(stdin)) {
        dwNumRead= fread(vectorptr(buffer), 1, buffer.size(), stdin);
        if (dwNumRead==0)
            break;

        DWORD dwNumWritten;
        if (!CeWriteFile( hDest, vectorptr(buffer), dwNumRead, &dwNumWritten, NULL))
        {
            if (verbose>0) ceerror("Writing WCE:%s", dstfile.c_str());
            bRes= false;
            break;
        }
    }

    CeCloseHandle( hDest);

    return bRes;
}
bool CopyCeFileToWin32File(const std::string& src, const std::string& dstfile)
{
    if (!g_allowOverwrite && isWin32File(dstfile))
    {
        if (verbose>1) debug("destination file already exists - specify -f to overwrite\n");
        return false;
    }
    HANDLE hSrc = CeCreateFile( ToWString(src).c_str(), GENERIC_READ, FILE_SHARE_READ,
                NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (INVALID_HANDLE_VALUE == hSrc)
    {
        if (verbose>1) ceerror("WCE:%s", src.c_str());
        return false;
    }

    HANDLE hDest = CreateFile( dstfile.c_str(), GENERIC_WRITE, FILE_SHARE_READ,
                NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
    if (INVALID_HANDLE_VALUE == hDest )
    {
        if (verbose>1) error("%s", dstfile.c_str());
        return false;
    }

    if (verbose>2) debug("Copying WCE:%hs to %hs\n", src.c_str(), dstfile.c_str());

    ByteVector buffer;  buffer.resize(65536);

    bool bRes= true;

    DWORD dwNumRead;
    do
    {
        if (!CeReadFile( hSrc, vectorptr(buffer), buffer.size(), &dwNumRead, NULL))
        {
            if (verbose>0) ceerror("reading WCE:%s", src.c_str());
            bRes= false;
            break;
        }

        DWORD dwNumWritten;
        if (!WriteFile( hDest, vectorptr(buffer), dwNumRead, &dwNumWritten, NULL))
        {
            if (verbose>0) error("Writing %s", dstfile.c_str());
            bRes= false;
            break;
        }
    } while (dwNumRead);

    CloseHandle( hDest);
    CeCloseHandle (hSrc);

    return bRes;
}
bool PrintCeFile(const std::string& src)
{
    HANDLE hSrc = CeCreateFile( ToWString(src).c_str(), GENERIC_READ, FILE_SHARE_READ,
                NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (INVALID_HANDLE_VALUE == hSrc)
    {
        if (verbose>1) ceerror("WCE:%s", src.c_str());
        return false;
    }

    if (verbose>1) debug("Printing WCE:%hs\n", src.c_str());

    ByteVector buffer;  buffer.resize(65536);

    bool bRes= true;

    DWORD dwNumRead;
    do
    {
        if (!CeReadFile( hSrc, vectorptr(buffer), buffer.size(), &dwNumRead, NULL))
        {
            if (verbose>0) ceerror("Reading WCE:%s", src.c_str());
            bRes= false;
            break;
        }
        fwrite(vectorptr(buffer), 1, dwNumRead, stdout);
    } while (dwNumRead);

    CeCloseHandle (hSrc);

    printf("\n");

    return bRes;
}

bool CopyWin32FileToCeDirectory(const std::string& src, const std::string& dstdir)
{
    size_t lastslash= src.find_last_of("\\/");
    if (lastslash==src.npos)
        return CopyWin32FileToCeFile(src, concat_path(dstdir, src));
    else
        return CopyWin32FileToCeFile(src, concat_path(dstdir, src.substr(lastslash+1)));
}
bool CopyCeFileToWin32Directory(const std::string& src, const std::string& dstdir)
{
    size_t lastslash= src.find_last_of("\\/");
    if (lastslash==src.npos)
        return CopyCeFileToWin32File(src, concat_path(dstdir, src));
    else                        
        return CopyCeFileToWin32File(src, concat_path(dstdir, src.substr(lastslash+1)));
}


bool CopyWin32ToCe(const StringList& args, const std::string& dst)
{
    if (isCeDirectory(dst)) {
        for (size_t i=0 ; i<args.size() ; i++) {
            std::string src= args[i];
            if (hasWildCards(src)) {
                StringList files;
                if (!Win32FileGlob(src, files))
                {
                    if (verbose>0) error("%s\n", src.c_str());
                    return false;
                }
                for (size_t j= 0 ; j<files.size() ; j++) {
                    CopyWin32FileToCeDirectory(files[j], dst);
                }
            }
            else {
                CopyWin32FileToCeDirectory(src, dst);
            }
        }
    }
    else {
        if (args.size()>1)
        {
            if (verbose>0) debug("multiple source files: target must be an existing directory\n");
            return false;
        }
        std::string src= args[0];
        if (hasWildCards(src))
        {
            StringList files;
            if (!Win32FileGlob(src, files))
            {
                if (verbose>0) error("%s\n", src.c_str());
                return false;
            }
            if (files.size()>1)
            {
                if (verbose>0) debug("source is wildcard to multiple files: target must be an existing directory\n");
                return false;
            }
            src= files[0];
        }
        CopyWin32FileToCeFile(src, dst);
    }
    return true;
}

bool CopyCeToWin32(const StringList& args, const std::string& dst)
{
    if (dst == "-") {
        // output all to stdout
        for (size_t i=0 ; i<args.size() ; i++) {
            std::string src= args[i];
            if (isCeDirectory(src))
                src= concat_path(src, "*");
            if (hasWildCards(src)) {
                StringList files, dirs;
                if (!CeFileGlob(src, files, dirs))
                {
                    if (verbose>0) error("WCE:%s\n", src.c_str());
                    return false;
                }
                for (size_t j= 0 ; j<files.size() ; j++) {
                    PrintCeFile(files[j]);
                }
                if (g_bRecurse) {
                    for (size_t j= 0 ; j<dirs.size() ; j++) {
                        StringList oneitem;
                        oneitem.push_back(concat_path(dirs[j], "*"));
                        CopyCeToWin32(oneitem, dst);
                    }
                }
            }
            else {
                PrintCeFile(src);
            }
        }
    }
    else if (isWin32Directory(dst)) {
        for (size_t i=0 ; i<args.size() ; i++) {
            std::string src= args[i];
            if (isCeDirectory(src))
                src= concat_path(src, "*");
            if (hasWildCards(src)) {
                StringList files, dirs;
                if (!CeFileGlob(src, files, dirs))
                {
                    if (verbose>0) error("WCE:%s\n", src.c_str());
                    return false;
                }
                for (size_t j= 0 ; j<files.size() ; j++) {
                    CopyCeFileToWin32Directory(files[j], dst);
                }
                if (g_bRecurse) {
                    for (size_t j= 0 ; j<dirs.size() ; j++) {
                        StringList oneitem;
                        oneitem.push_back(concat_path(dirs[j], "*"));
                        CreateDirectory(concat_path(dst, GetBaseName(dirs[j])).c_str(), NULL);
                        CopyCeToWin32(oneitem, concat_path(dst, GetBaseName(dirs[j])));
                    }
                }
            }
            else {
                CopyCeFileToWin32Directory(src, dst);
            }
        }
    }
    else {
        if (args.size()>1)
        {
            if (verbose>0) debug("multiple source files: target must be an existing directory\n");
            return false;
        }
        std::string src= args[0];
        if (isCeDirectory(src))
            src= concat_path(src, "*");
        if (hasWildCards(src))
        {
            StringList files, dirs;
            if (!CeFileGlob(src, files, dirs))
            {
                if (verbose>0) error("WCE:%s\n", src.c_str());
                return false;
            }
            if (files.size()>1)
            {
                if (verbose>0) debug("source is wildcard to multiple files: target must be an existing directory\n");
                return false;
            }
            src= files[0];
        }
        CopyCeFileToWin32File(src, dst);
    }
    return true;
}
void usage(const std::string& cmd)
{
    if (cmd.substr(0,4)=="pput")
        debug("Usage: pput [-f] <localfile(s)> <devicefile|path>\n");
    else if (cmd.substr(0,4)=="pget")
        debug("Usage: pget [-f] <deviefile(s)> <localfile|path>\n");
    debug("     -f : overwrite existing files\n");
    debug("     -r : recursive copy\n");
    debug("     -q : less verbose output\n");
    debug("     -v : more verbose output\n");
    debug("       source files may contain wildcards\n");
}

int main(int argc, char *argv[])
{
    std::string cmd= tolower(GetBaseName(argv[0]));

    DebugStdOut();

    StringList args;

    for (int i=1 ; i<argc ; i++)
    {
        if (argv[i][0]=='-') switch(argv[i][1])
        {
            case 'f': g_allowOverwrite= true; break;
            case 'r': g_bRecurse= true; break;
            case 'c': g_useStdout= true; break;
            case 'v': verbose++; break;
            case 'q': verbose--; break;
            default:
                usage(cmd);
                return 1;
        }
        else {
            args.push_back(argv[i]);
        }
    }

    if (FAILED(CeRapiInit()))
    {
        error("rapi init Failed");
        return 1;
    }

    if (cmd.substr(0,4)=="pput") {
        if (g_useStdout) {
            if (args.size()!=1)
            {
                usage(cmd);
                return 1;
            }
            CopyStdinToCeFile(args.back());
        }
        else if (args.size()<2)
        {
            usage(cmd);
            return 1;
        }
        else {
            std::string dst= args.back();  args.pop_back();

            CopyWin32ToCe(args, dst);
        }
    }
    else if (cmd.substr(0,4)=="pget") {
        std::string dst;
        if (args.size()==0)
        {
            usage(cmd);
            return 1;
        }
        else if (g_useStdout) {
            dst= "-";
        }
        else if (args.size()==1)
        {
            dst= ".";
        }
        else {
            dst= args.back();  args.pop_back();
        }
        CopyCeToWin32(args, dst);
    }
    else {
        debug("program must be called either 'pput' or 'pget'\n");
        return 1;
    }

    CeRapiUninit();

    return 0;
}


